"""
The HTTP Manager has primary responsibility over mapping HTTP requests to
specific Python functions (the endpoints), and converting the responses of
those functions to HTTP responses. Its methods are designed to be invoked
solely through decorators, thus completely separating business logic from
HTTP logic.

Specifically, the HTTP Manager currently does three things:
1. Uses the Flask library to correctly set up the HTTP request to Python
   function mappings.
2. Implements all of the logic for mapping Transiter service layer responses
   (which are typically Python dicts) to 3 tuples
   (string response, HTTP code, HTTP headers) that are then used by Flask to
   send the correct HTTP response.
3. Converts service layer Links to HTTP URLs. This is done with the assistance
   of the link_target decorator, which is used to identify which endpoints
   correspond to which Links.
"""
import dataclasses
import decimal
import enum
import inspect
import json
import logging
from datetime import date, datetime

import flask
from decorator import decorator

from transiter import exceptions
from transiter.services import views

logger = logging.getLogger(__name__)


class HttpStatus(enum.IntEnum):
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NO_CONTENT = 204
    BAD_REQUEST = 400
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    SERVICE_UNAVAILABLE = 503


_exception_type_to_http_status = {
    exceptions.AccessDenied: HttpStatus.FORBIDDEN,
    exceptions.UnexpectedError: HttpStatus.INTERNAL_SERVER_ERROR,
    exceptions.IdNotFoundError: HttpStatus.NOT_FOUND,
    exceptions.PageNotFound: HttpStatus.NOT_FOUND,
    exceptions.MethodNotAllowed: HttpStatus.METHOD_NOT_ALLOWED,
    exceptions.InstallError: HttpStatus.INTERNAL_SERVER_ERROR,
    exceptions.InvalidInput: HttpStatus.BAD_REQUEST,
    exceptions.InvalidSystemConfigFile: HttpStatus.BAD_REQUEST,
    exceptions.InvalidPermissionsLevelInRequest: HttpStatus.INTERNAL_SERVER_ERROR,
    exceptions.InternalDocumentationMisconfigured: HttpStatus.SERVICE_UNAVAILABLE,
}


class HttpMethod(enum.Enum):
    DELETE = "DELETE"
    GET = "GET"
    POST = "POST"
    PUT = "PUT"


def http_endpoint(
    flask_entity,
    flask_rule,
    method: HttpMethod = HttpMethod.GET,
    returns_json_response: bool = True,
):
    """
    Decorator factory used to easily register a Transiter HTTP endpoint.

    :param flask_entity: either the Flask app or a Flask blueprint
    :param flask_rule: the URL relative to the Flask entity
    :param method: which HTTP method this endpoint uses
    :param returns_json_response: whether to format the response as JSON and apply
        appropriate HTTP headers
    """
    decorators = [
        flask_entity.route(flask_rule + "/", methods=[method.value]),
        flask_entity.route(flask_rule, methods=[method.value]),
    ]
    if returns_json_response:
        decorators.append(_json_response)

    def composed_decorator(func):
        for decorator_ in reversed(decorators):
            func = decorator_(func)
        return func

    return composed_decorator


@decorator
def _json_response(func, *args, **kwargs):
    response = func(*args, **kwargs)
    status = HttpStatus.OK
    if (
        isinstance(response, tuple)
        and len(response) == 2
        and isinstance(response[1], HttpStatus)
    ):
        response, status = response
    return flask.Response(
        response=_convert_to_json_str(response),
        status=status,
        content_type="application/json",
    )


_link_type_to_target = {}


@dataclasses.dataclass
class ViewTarget:
    target_str: str
    target_param_to_view_param: dict


_view_to_target = {}


def link_target(link_type, view_params=None):
    """
    This decorator factory is used to identify a HTTP endpoint with a given
    view class. It's used as follows:

    .. code-block:: python

        @link_target(views.Stop, ["id"])
        def http_endpoint_to_link(stop_id):
            pass

    :param link_type: the view type
    :param view_params: a list of string field names in the view whose values would be
     passed to the endpoint to get the response for that view object.
    :return: the factory returns a decorator to be applied to the endpoint.
    """

    def views_decorator(func):
        flask_endpoint = "{}.{}".format(func.__module__, func.__name__)
        target_param_to_view_param = {}
        if view_params is not None:
            for i, target_param in enumerate(inspect.signature(func).parameters.keys()):
                target_param_to_view_param[target_param] = view_params[i]
        _view_to_target[link_type] = ViewTarget(
            flask_endpoint, target_param_to_view_param
        )
        return func

    return views_decorator


def get_url_parameters(expected_keys, error_if_extra_keys=True):
    all_request_args = flask.request.args
    if error_if_extra_keys:
        extra_keys = set(all_request_args.keys()) - set(expected_keys)
        if len(extra_keys) > 0:
            raise exceptions.InvalidInput(
                "Unknown URL parameters: {}. Valid URL parameters for this endpoint: {}".format(
                    extra_keys, expected_keys
                )
            )
    return {key: all_request_args.get(key) for key in expected_keys}


def is_sync_request():
    return (
        get_url_parameters(["sync"], error_if_extra_keys=False).get("sync", "false")
        == "true"
    )


def convert_exception_to_error_response(exception):
    # noinspection PyBroadException
    try:
        logger.debug("Exception", exc_info=True)
        return flask.Response(
            response=_convert_to_json_str(exception.response()),
            status=_exception_type_to_http_status[type(exception)],
            content_type="application/json",
        )
    except Exception:
        logger.exception("Unexpected exception in processing HTTP request.")
        return flask.Response(response="", status=HttpStatus.INTERNAL_SERVER_ERROR)


def _convert_to_json_str(data):
    """
    Convert a server layer response to a JSON string.
    """
    return json.dumps(
        data, indent=2, separators=(",", ": "), default=_transiter_json_serializer
    )


def _build_href(view: views.View):
    if type(view) not in _view_to_target:
        return None
    view_target = _view_to_target[type(view)]
    custom_host = flask.request.headers.get("X-Transiter-Host")
    kwargs = {
        key: getattr(view, value)
        for key, value in view_target.target_param_to_view_param.items()
    }
    if custom_host is not None:
        return custom_host + flask.url_for(
            view_target.target_str, _external=False, **kwargs
        )
    return flask.url_for(view_target.target_str, _external=True, **kwargs)


def _transiter_json_serializer(obj):
    """
    This is custom Transiter JSON serializer for objects that are not
    serializable by default.
    """
    if isinstance(obj, views.View):
        base = obj.to_dict()
        href = _build_href(obj)
        if href is not None:
            base["href"] = href
        return base

    if isinstance(obj, (datetime, date)):
        return obj.timestamp()

    if isinstance(obj, enum.Enum):
        return obj.name

    if isinstance(obj, decimal.Decimal):
        return str(obj)

    raise TypeError("Type {} not serializable".format(type(obj)))



"""
The route service is used to retrieve data about routes.
"""
import typing

from transiter import exceptions, models
from transiter.data import dbconnection
from transiter.data.dams import routedam, systemdam
from transiter.models import Alert
from transiter.services import views
from transiter.services.servicemap import servicemapmanager


@dbconnection.unit_of_work
def list_all_in_system(system_id) -> typing.List[views.Route]:
    """
    Get data on all routes in a system.
    """
    system = systemdam.get_by_id(system_id, only_return_active=True)
    if system is None:
        raise exceptions.IdNotFoundError(models.System, system_id=system_id)
    response = []
    routes = list(routedam.list_all_in_system(system_id))
    route_pk_to_status = _construct_route_pk_to_status_map(route.pk for route in routes)
    for route in routes:
        route_response = views.Route.from_model(route)
        route_response.status = route_pk_to_status[route.pk]
        response.append(route_response)
    return response


@dbconnection.unit_of_work
def get_in_system_by_id(system_id, route_id) -> views.RouteLarge:
    """
    Get data for a specific route in a specific system.
    """
    route = routedam.get_in_system_by_id(system_id, route_id)
    if route is None:
        raise exceptions.IdNotFoundError(
            models.Route, system_id=system_id, route_id=route_id
        )
    periodicity = routedam.calculate_periodicity(route.pk)
    if periodicity is not None:
        periodicity = int(periodicity / 6) / 10
    return views.RouteLarge.from_model(
        route,
        _construct_route_status(route.pk),
        periodicity,
        list(map(views.AlertLarge.from_model, route.route_statuses)),
        servicemapmanager.build_route_service_maps_response(route.pk),
    )


Status = views.Route.Status


def _construct_route_status(route_pk):
    """
    Construct the status for a specific route.

    See _construct_route_pk_to_status_map for documentation on how this is
    constructed.
    """
    return _construct_route_pk_to_status_map([route_pk])[route_pk]


def _construct_route_pk_to_status_map(route_pks_iter):
    """
    Construct the statuses for multiple routes.

    The algorithm constructs the status by first examining the highest priority
    alerts for a specific route. If there are alerts then,
     * If one of these alerts has the effect SIGNIFICANT_DELAYs, then the status
       of the route is DELAYS.
     * Otherwise, if one of the alerts has planned false, then the status is
       UNPLANNED_SERVICE_CHANGE.
     * Otherwise, the status is PLANNED_SERVICE_CHANGE.

    If there are now no alerts then it is checked to see if there are active
    trips for this route.
     * If so, the status is GOOD_SERVICE.
     * Otherwise, the status is NO_SERVICE.
    """
    route_pks = {route_pk for route_pk in route_pks_iter}

    route_pk_to_alerts = routedam.get_route_pk_to_highest_priority_alerts_map(route_pks)

    route_pk_to_status = {route_pk: Status.NO_SERVICE for route_pk in route_pks}
    for route_pk, alerts in route_pk_to_alerts.items():
        if len(alerts) == 0:
            continue
        causes = set(alert.cause for alert in alerts)
        effects = set(alert.effect for alert in alerts)
        if Alert.Effect.SIGNIFICANT_DELAYS in effects:
            route_pk_to_status[route_pk] = Status.DELAYS
        elif Alert.Cause.ACCIDENT in causes:
            route_pk_to_status[route_pk] = Status.UNPLANNED_SERVICE_CHANGE
        else:
            route_pk_to_status[route_pk] = Status.PLANNED_SERVICE_CHANGE
        route_pks.remove(route_pk)

    for route_pk in routedam.list_route_pks_with_current_service(route_pks):
        route_pk_to_status[route_pk] = Status.GOOD_SERVICE

    return route_pk_to_status



"""Add system update table

Revision ID: 055583d82811
Revises: 65146f234144
Create Date: 2020-03-24 11:14:28.985576

"""
import datetime

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "055583d82811"
down_revision = "65146f234144"
branch_labels = None
depends_on = None


def upgrade():
    # (1) Create the new table
    table = op.create_table(
        "system_update",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("system_pk", sa.Integer(), nullable=True),
        sa.Column(
            "status",
            sa.Enum(
                "SCHEDULED",
                "IN_PROGRESS",
                "SUCCESS",
                "FAILED",
                name="status",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.Column("status_message", sa.String(), nullable=True),
        sa.Column("total_duration", sa.Float(), nullable=True),
        sa.Column("scheduled_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("completed_at", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("config", sa.String(), nullable=True),
        sa.Column("config_template", sa.String(), nullable=True),
        sa.Column("config_parameters", sa.String(), nullable=True),
        sa.Column("config_source_url", sa.String(), nullable=True),
        sa.Column("transiter_version", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"]),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(
        "system_update_system_pk_system_update_pk_idx",
        "system_update",
        ["system_pk", "pk"],
        unique=False,
    )

    # (2) Migrate some of the data from the system table over
    statement = sa.sql.select(
        [sa.text("system.pk"), sa.text("system.raw_config"), sa.text("system.status")]
    ).select_from(sa.text("system"))
    data_dicts = []
    for system_pk, raw_config, system_status in op.get_bind().execute(statement):
        if system_status == "ACTIVE":
            status = "SUCCESS"
        else:
            status = "FAILED"
        data_dicts.append(
            {
                "system_pk": system_pk,
                "config_template": raw_config,
                "status": status,
                "status_message": "update data added as part of migration 0555",
                "scheduled_at": datetime.datetime.utcnow(),
                "completed_at": datetime.datetime.utcnow(),
                "transiter_version": "0.4.1",
            }
        )
    op.bulk_insert(table, data_dicts)

    # (3) Drop the old columns in the system table
    op.drop_column("system", "error_message")
    op.drop_column("system", "raw_config")

    # (4) Other operations as part of this branch
    op.add_column(
        "feed",
        sa.Column(
            "required_for_install", sa.Boolean(), nullable=False, server_default="false"
        ),
    )


def downgrade():
    op.drop_table("system_update")



"""
The config module is responsible for loading Transiter's global config at
runtime and providing it to modules that require it.
"""
import logging
import os

from distutils.util import strtobool

logger = logging.getLogger(__name__)


DB_DRIVER = "postgresql"
DB_HOST = "localhost"
DB_PORT = "5432"
DB_DATABASE = "transiter"
DB_USERNAME = "transiter"
DB_PASSWORD = "transiter"


SCHEDULER_HOST = "localhost"
SCHEDULER_PORT = "5000"

DOCUMENTATION_ENABLED = False
DOCUMENTATION_ROOT = "../../docs/site"


for env_variable_name, value in os.environ.items():
    prefix = "TRANSITER_"
    if env_variable_name[: len(prefix)] != prefix:
        continue
    variable_name = env_variable_name[len(prefix) :]
    if variable_name not in globals():
        logger.debug(
            "Skipping unknown Transiter environment setting '{}'.".format(variable_name)
        )
        continue
    if isinstance(globals()[variable_name], bool):
        value = bool(strtobool(value))
    logger.debug(
        "Setting Transiter environment variable '{}' to be '{}'.".format(
            variable_name, value
        )
    )
    globals()[variable_name] = value



from typing import Optional, NamedTuple, Dict, List

import sqlalchemy
import sqlalchemy.sql.expression as sql
from sqlalchemy.orm import selectinload, joinedload

from transiter import models
from transiter.data import dbconnection


def list_all_from_feed(feed_pk):
    session = dbconnection.get_session()
    query = session.query(models.Trip).filter(
        models.Trip.source_pk == models.FeedUpdate.pk,
        models.FeedUpdate.feed_pk == feed_pk,
    )
    return query.all()


def list_by_system_and_trip_ids(system_pk, trip_ids):
    if len(trip_ids) == 0:
        return []
    return (
        dbconnection.get_session()
        .query(models.Trip)
        .join(models.Route)
        .filter(models.Trip.id.in_(trip_ids))
        .filter(models.Route.system_pk == system_pk)
        .all()
    )


class StopTimeData(NamedTuple):
    pk: int
    stop_sequence: int
    stop_pk: int


def get_trip_pk_to_stop_time_data_list(feed_pk) -> Dict[int, List[StopTimeData]]:
    session = dbconnection.get_session()
    query = (
        session.query(
            models.TripStopTime.trip_pk,
            models.TripStopTime.pk,
            models.TripStopTime.stop_sequence,
            models.TripStopTime.stop_pk,
        )
        .join(models.Trip, models.Trip.pk == models.TripStopTime.trip_pk)
        .join(models.FeedUpdate, models.FeedUpdate.pk == models.Trip.source_pk)
        .filter(models.FeedUpdate.feed_pk == feed_pk)
        .order_by(models.TripStopTime.trip_pk, models.TripStopTime.stop_sequence)
    )
    trip_pk_to_stop_time_data_list = {}
    for (trip_pk, stop_time_pk, stop_sequence, stop_pk,) in query.all():
        if trip_pk not in trip_pk_to_stop_time_data_list:
            trip_pk_to_stop_time_data_list[trip_pk] = []
        trip_pk_to_stop_time_data_list[trip_pk].append(
            StopTimeData(pk=stop_time_pk, stop_sequence=stop_sequence, stop_pk=stop_pk)
        )
    return trip_pk_to_stop_time_data_list


def list_all_in_route_by_pk(route_pk):
    """
    List all of the Trips in a route.

    :param route_pk: the route's PK
    :return: list of Trips
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.Trip)
        .filter(models.Trip.route_pk == route_pk)
        .options(selectinload(models.Trip.stop_times))
    )
    return query.all()


def get_in_route_by_id(system_id, route_id, trip_id) -> Optional[models.Trip]:
    """
    Get a trip using IDs.

    This method greedily loads the trip's route, its stop times, and the stops
    associated to those stop times.
    """
    session = dbconnection.get_session()
    return (
        session.query(models.Trip)
        .join(models.Route, models.Route.pk == models.Trip.route_pk)
        .join(models.System, models.System.pk == models.Route.system_pk)
        .filter(models.System.id == system_id)
        .filter(models.Route.id == route_id)
        .filter(models.Trip.id == trip_id)
        .options(joinedload(models.Trip.route))
        .options(selectinload(models.Trip.stop_times))
        .options(selectinload(models.Trip.stop_times, models.TripStopTime.stop))
        .one_or_none()
    )


def get_trip_pk_to_last_stop_map(trip_pks):
    """
    Get the map to trip PK to the last stop for the trip.

    :param trip_pks: the trip PKs to build the map for
    :return: trip_pk to Stop
    """
    session = dbconnection.get_session()

    sub_query = (
        session.query(
            models.TripStopTime.trip_pk,
            sqlalchemy.func.max(models.TripStopTime.stop_sequence),
        )
        .group_by(models.TripStopTime.trip_pk)
        .filter(models.TripStopTime.trip_pk.in_(trip_pks))
    )
    query = (
        session.query(models.TripStopTime.trip_pk, models.Stop)
        .filter(models.TripStopTime.stop_pk == models.Stop.pk)
        .filter(
            sqlalchemy.tuple_(
                models.TripStopTime.trip_pk, models.TripStopTime.stop_sequence
            ).in_(sub_query)
        )
    )
    trip_pk_to_last_stop = {trip_pk: None for trip_pk in trip_pks}
    for trip_pk, last_stop in query:
        trip_pk_to_last_stop[trip_pk] = last_stop
    return trip_pk_to_last_stop


def get_trip_pk_to_path_map(route_pk):
    """
    Get a map of trip PK to the path of the trip for every trip in a route.

    The path here is a list of stop PKs.

    :param route_pk: the route's PK
    :return: map described above
    """
    statement = (
        sql.select([models.TripStopTime.trip_pk, models.TripStopTime.stop_pk])
        .select_from(sql.join(models.TripStopTime, models.Trip))
        .where(models.Trip.route_pk == route_pk)
        .order_by(models.TripStopTime.trip_pk, models.TripStopTime.stop_sequence)
    )
    session = dbconnection.get_session()
    trip_pk_to_stop_pks = {}
    for trip_pk, stop_pk in session.execute(statement):
        if trip_pk not in trip_pk_to_stop_pks:
            trip_pk_to_stop_pks[trip_pk] = []
        trip_pk_to_stop_pks[trip_pk].append(stop_pk)
    return trip_pk_to_stop_pks



from sqlalchemy import Column, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from .base import Base


class ServiceMap(Base):
    __tablename__ = "service_map"

    pk = Column(Integer, primary_key=True)
    route_pk = Column(Integer, ForeignKey("route.pk"), index=True, nullable=False)
    group_pk = Column(Integer, ForeignKey("service_map_group.pk"), nullable=False)

    group = relationship("ServiceMapGroup", back_populates="maps")
    route = relationship(
        "Route", foreign_keys=[route_pk], back_populates="service_maps"
    )
    vertices = relationship(
        "ServiceMapVertex",
        back_populates="map",
        order_by="ServiceMapVertex.position",
        cascade="all, delete-orphan",
    )

    __table_args__ = (UniqueConstraint(route_pk, group_pk),)



from sqlalchemy import inspect
from sqlalchemy.ext.declarative import declarative_base


class _BaseModel:
    """
    Basic model that all Transiter models inherit from.

    This class defines an equals method for use in testing.
    """

    # The equals method is designed for testing purposes only.
    def __eq__(self, other):
        for column in inspect(type(self)).columns.keys():
            if getattr(self, column) != getattr(other, column):
                print(
                    'Values for attribute "{}" don\'t match: "{}" != "{}"'.format(
                        column, str(getattr(self, column)), str(getattr(other, column))
                    )
                )
                return False
        return True

    # The __repr__ method is designed for testing purposes only.
    def __repr__(self):  # pragma: no cover
        attributes = []
        for column in inspect(type(self)).columns.keys():
            attributes.append("{}={}".format(column, getattr(self, column)))
        return "{}({})".format(type(self).__name__, ", ".join(attributes))

    # This is a hack to get around a problem in SQL Alchemy where the hash is evaluated
    # when setting models.Alert.routes = [...].
    def __hash__(self):
        return id(self)


Base = declarative_base(cls=_BaseModel)



"""
This module provides some abstract methods to remove code duplication in the DAMs.
"""
import typing
from typing import Dict, Iterable

from sqlalchemy.orm import joinedload

from transiter import models
from transiter.data import dbconnection


def create(DbEntity: models.Base, entity=None):
    """
    Create a database entity.

    :param DbEntity: the entity's type
    :param entity: optionally an instance of this type that will instead be added
        to the session
    :return: the entity, which is in the session
    """
    session = dbconnection.get_session()
    if entity is None:
        entity = DbEntity()
    session.add(entity)
    return entity


def list_all(DbEntity: models.Base, order_by_field=None):
    """
    List all entities of a certain type.

    :param DbEntity: the entities' type
    :param order_by_field: field or order the results be
    :return: list of entities of type DbEntity
    """
    session = dbconnection.get_session()
    query = session.query(DbEntity)
    if order_by_field is not None:
        query = query.order_by(order_by_field)
    return query.all()


def get_by_id(DbEntity: models.Base, id_):
    """
    Get an entity by its ID.

    :param DbEntity: the entity's type
    :param id_: the entity's ID
    :return: the entity, if it exists in the DB, or None otherwise
    """
    session = dbconnection.get_session()
    return session.query(DbEntity).filter(DbEntity.id == id_).one_or_none()


def list_all_in_system(DbEntity: models.Base, system_id, order_by_field=None):
    """
    List all entities of a certain type that are in a given system. Note this method
    only works with entities that are direct children of the system.

    :param DbEntity: the entity's type
    :param system_id: the system's ID
    :param order_by_field: optional field to order the results by
    :return: list of entities of type DbEntity
    """
    session = dbconnection.get_session()
    query = (
        session.query(DbEntity)
        .filter(DbEntity.system_pk == models.System.pk)
        .filter(models.System.id == system_id)
    )
    if order_by_field is not None:
        query = query.order_by(order_by_field)
    return query.all()


def get_in_system_by_id(DbEntity: models.Base, system_id, id_):
    """
    Get an entity of a certain type that is in a given system. Note this method
    only works with entities that are direct children of the system.

    :param DbEntity: the entity's type
    :param system_id: the system's ID
    :param id_: the entity's ID
    :return: list of entities of type DbEntity
    """
    session = dbconnection.get_session()
    return (
        session.query(DbEntity)
        .filter(DbEntity.system_pk == models.System.pk)
        .filter(models.System.id == system_id)
        .filter(DbEntity.id == id_)
        .options(joinedload(DbEntity.system))
        .one_or_none()
    )


def get_id_to_pk_map(
    DbEntity: models.Base, system_pk: int = None, ids: Iterable[str] = None
) -> Dict[str, int]:
    """
    Get an map of entity ID to entity PK for all entities of a given type in a system.
    Note this method only works with entities that are direct children of the system.
    """
    if ids is not None:
        id_to_pk = {id_: None for id_ in ids}
    else:
        id_to_pk = {}
    session = dbconnection.get_session()
    query = session.query(DbEntity.id, DbEntity.pk)
    if system_pk is not None:
        query = query.filter(DbEntity.system_pk == system_pk)
    if ids is not None:
        query = query.filter(DbEntity.id.in_(ids))
    for (id_, pk) in query.all():
        id_to_pk[id_] = pk
    return id_to_pk


# DbEntity is a class
# noinspection PyPep8Naming
def get_id_to_pk_map_by_feed_pk(DbEntity: typing.Type[models.Base], feed_pk):
    id_to_pk = {}
    session = dbconnection.get_session()
    query = (
        session.query(DbEntity.id, DbEntity.pk)
        .join(models.FeedUpdate, DbEntity.source_pk == models.FeedUpdate.pk)
        .filter(models.FeedUpdate.feed_pk == feed_pk)
    )
    for (id_, pk) in query.all():
        id_to_pk[id_] = pk
    return id_to_pk


# DbEntity is a class
# noinspection PyPep8Naming
def delete_stale_entities(
    DbEntity: typing.Type[models.Base], feed_update: models.FeedUpdate
):
    session = dbconnection.get_session()
    (
        session.query(DbEntity)
        .filter(DbEntity.source_pk == models.FeedUpdate.pk)
        .filter(models.FeedUpdate.feed_pk == feed_update.feed_pk)
        .filter(models.FeedUpdate.pk != feed_update.pk)
    ).delete(synchronize_session=False)


# DbEntity is a class
# noinspection PyPep8Naming
def list_stale_entities(
    DbEntity: typing.Type[models.Base], feed_update: models.FeedUpdate
):
    session = dbconnection.get_session()
    query = (
        session.query(DbEntity)
        .join(models.FeedUpdate, DbEntity.source_pk == models.FeedUpdate.pk)
        .filter(models.FeedUpdate.feed_pk == feed_update.feed_pk)
        .filter(models.FeedUpdate.pk != feed_update.pk)
    )
    return query.all()



"""
Initial database setup.

Revision ID: 14fa42915420
Revises:
Create Date: 2020-03-08 20:18:55.140759

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "14fa42915420"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "system",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column(
            "status",
            sa.Enum(
                "SCHEDULED",
                "INSTALLING",
                "ACTIVE",
                "INSTALL_FAILED",
                "DELETING",
                name="systemstatus",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.Column("error_message", sa.String(), nullable=True),
        sa.Column("timezone", sa.String(), nullable=True),
        sa.Column("raw_config", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(op.f("ix_system_id"), "system", ["id"], unique=True)
    op.create_table(
        "feed",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("system_pk", sa.Integer(), nullable=True),
        sa.Column(
            "built_in_parser",
            sa.Enum(
                "GTFS_STATIC", "GTFS_REALTIME", name="builtinparser", native_enum=False
            ),
            nullable=True,
        ),
        sa.Column("custom_parser", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column("headers", sa.String(), nullable=True),
        sa.Column("auto_update_on", sa.Boolean(), nullable=True),
        sa.Column("auto_update_period", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("system_pk", "id"),
    )
    op.create_table(
        "service_map_group",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("system_pk", sa.Integer(), nullable=False),
        sa.Column(
            "source",
            sa.Enum("SCHEDULE", "REALTIME", name="servicemapsource", native_enum=False),
            nullable=False,
        ),
        sa.Column("conditions", sa.String(), nullable=True),
        sa.Column("threshold", sa.Float(), nullable=False),
        sa.Column("use_for_routes_at_stop", sa.Boolean(), nullable=False),
        sa.Column("use_for_stops_in_route", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_table(
        "feed_update",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("feed_pk", sa.Integer(), nullable=True),
        sa.Column(
            "update_type",
            sa.Enum("REGULAR", "FLUSH", name="update_type", native_enum=False),
            nullable=False,
        ),
        sa.Column(
            "status",
            sa.Enum(
                "SCHEDULED",
                "IN_PROGRESS",
                "SUCCESS",
                "FAILURE",
                name="status",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column(
            "explanation",
            sa.Enum(
                "UPDATED",
                "NOT_NEEDED",
                "PARSE_ERROR",
                "DOWNLOAD_ERROR",
                "INVALID_PARSER",
                "EMPTY_FEED",
                "SYNC_ERROR",
                "UNEXPECTED_ERROR",
                name="explanation",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column("failure_message", sa.String(), nullable=True),
        sa.Column("raw_data_hash", sa.String(), nullable=True),
        sa.Column("content_length", sa.Integer(), nullable=True),
        sa.Column("execution_duration", sa.Float(), nullable=True),
        sa.Column(
            "last_action_time",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("feed_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.ForeignKeyConstraint(["feed_pk"], ["feed.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(
        "feed_updates_last_successful_idx",
        "feed_update",
        ["feed_pk", "last_action_time", "status"],
        unique=False,
    )
    op.create_index(
        op.f("ix_feed_update_last_action_time"),
        "feed_update",
        ["last_action_time"],
        unique=False,
    )
    op.create_table(
        "alert",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("header", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column(
            "cause",
            sa.Enum(
                "UNKNOWN_CAUSE",
                "OTHER_CAUSE",
                "TECHNICAL_PROBLEM",
                "STRIKE",
                "DEMONSTRATION",
                "ACCIDENT",
                "HOLIDAY",
                "WEATHER",
                "MAINTENANCE",
                "CONSTRUCTION",
                "POLICE_ACTIVITY",
                "MEDICAL_EMERGENCY",
                name="cause",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column(
            "effect",
            sa.Enum(
                "NO_SERVICE",
                "REDUCED_SERVICE",
                "SIGNIFICANT_DELAYS",
                "DETOUR",
                "ADDITIONAL_SERVICE",
                "MODIFIED_SERVICE",
                "OTHER_EFFECT",
                "UNKNOWN_EFFECT",
                "STOP_MOVED",
                name="effect",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column("priority", sa.Integer(), nullable=True),
        sa.Column("start_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("end_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("creation_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("system_pk", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(op.f("ix_alert_source_pk"), "alert", ["source_pk"], unique=False)
    op.create_index(op.f("ix_alert_system_pk"), "alert", ["system_pk"], unique=False)
    op.create_table(
        "route",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("system_pk", sa.Integer(), nullable=False),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("color", sa.String(), nullable=True),
        sa.Column("text_color", sa.String(), nullable=True),
        sa.Column("short_name", sa.String(), nullable=True),
        sa.Column("long_name", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column(
            "type",
            sa.Enum(
                "LIGHT_RAIL",
                "SUBWAY",
                "RAIL",
                "BUS",
                "FERRY",
                "CABLE_CAR",
                "GONDOLA",
                "FUNICULAR",
                name="route_type",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column("sort_order", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("system_pk", "id"),
    )
    op.create_index(op.f("ix_route_source_pk"), "route", ["source_pk"], unique=False)
    op.create_table(
        "scheduled_service",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("system_pk", sa.Integer(), nullable=True),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("monday", sa.Boolean(), nullable=True),
        sa.Column("tuesday", sa.Boolean(), nullable=True),
        sa.Column("wednesday", sa.Boolean(), nullable=True),
        sa.Column("thursday", sa.Boolean(), nullable=True),
        sa.Column("friday", sa.Boolean(), nullable=True),
        sa.Column("saturday", sa.Boolean(), nullable=True),
        sa.Column("sunday", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(
        op.f("ix_scheduled_service_source_pk"),
        "scheduled_service",
        ["source_pk"],
        unique=False,
    )
    op.create_index(
        op.f("ix_scheduled_service_system_pk"),
        "scheduled_service",
        ["system_pk"],
        unique=False,
    )
    op.create_table(
        "stop",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("system_pk", sa.Integer(), nullable=False),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("parent_stop_pk", sa.Integer(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("longitude", sa.Numeric(precision=9, scale=6), nullable=True),
        sa.Column("latitude", sa.Numeric(precision=9, scale=6), nullable=True),
        sa.Column("url", sa.String(), nullable=True),
        sa.Column("is_station", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(["parent_stop_pk"], ["stop.pk"],),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.ForeignKeyConstraint(["system_pk"], ["system.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("system_pk", "id"),
    )
    op.create_index(
        op.f("ix_stop_parent_stop_pk"), "stop", ["parent_stop_pk"], unique=False
    )
    op.create_index(op.f("ix_stop_source_pk"), "stop", ["source_pk"], unique=False)
    op.create_table(
        "alert_route",
        sa.Column("alert_pk", sa.Integer(), nullable=True),
        sa.Column("route_pk", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["alert_pk"], ["alert.pk"],),
        sa.ForeignKeyConstraint(["route_pk"], ["route.pk"],),
    )
    op.create_index(
        op.f("ix_alert_route_route_pk"), "alert_route", ["route_pk"], unique=False
    )
    op.create_table(
        "direction_name_rule",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("stop_pk", sa.Integer(), nullable=True),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("priority", sa.Integer(), nullable=True),
        sa.Column("direction_id", sa.Boolean(), nullable=True),
        sa.Column("track", sa.String(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.ForeignKeyConstraint(["stop_pk"], ["stop.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(
        "direction_name_rule_stop_pk_priority_idx",
        "direction_name_rule",
        ["stop_pk", "priority"],
        unique=False,
    )
    op.create_index(
        op.f("ix_direction_name_rule_source_pk"),
        "direction_name_rule",
        ["source_pk"],
        unique=False,
    )
    op.create_table(
        "scheduled_trip",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("route_pk", sa.Integer(), nullable=False),
        sa.Column("service_pk", sa.Integer(), nullable=False),
        sa.Column("direction_id", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(["route_pk"], ["route.pk"],),
        sa.ForeignKeyConstraint(["service_pk"], ["scheduled_service.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_table(
        "service_map",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("route_pk", sa.Integer(), nullable=False),
        sa.Column("group_pk", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["group_pk"], ["service_map_group.pk"],),
        sa.ForeignKeyConstraint(["route_pk"], ["route.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("route_pk", "group_pk"),
    )
    op.create_index(
        op.f("ix_service_map_route_pk"), "service_map", ["route_pk"], unique=False
    )
    op.create_table(
        "trip",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(), nullable=True),
        sa.Column("route_pk", sa.Integer(), nullable=False),
        sa.Column("source_pk", sa.Integer(), nullable=True),
        sa.Column("direction_id", sa.Boolean(), nullable=True),
        sa.Column("start_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("last_update_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("vehicle_id", sa.String(), nullable=True),
        sa.Column(
            "current_status",
            sa.Enum(
                "SCHEDULED",
                "INCOMING_AT",
                "STOPPED_AT",
                "IN_TRANSIT_TO",
                name="tripstatus",
                native_enum=False,
            ),
            nullable=True,
        ),
        sa.Column("current_stop_sequence", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["route_pk"], ["route.pk"],),
        sa.ForeignKeyConstraint(["source_pk"], ["feed_update.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("route_pk", "id"),
    )
    op.create_index(op.f("ix_trip_source_pk"), "trip", ["source_pk"], unique=False)
    op.create_table(
        "scheduled_trip_stop_time",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("trip_pk", sa.Integer(), nullable=False),
        sa.Column("stop_pk", sa.Integer(), nullable=False),
        sa.Column("arrival_time", sa.Time(), nullable=True),
        sa.Column("departure_time", sa.Time(), nullable=True),
        sa.Column("stop_sequence", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["stop_pk"], ["stop.pk"],),
        sa.ForeignKeyConstraint(["trip_pk"], ["scheduled_trip.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("trip_pk", "stop_sequence"),
    )
    op.create_index(
        "scheduled_trip_stop_time_trip_pk_departure_time_idx",
        "scheduled_trip_stop_time",
        ["trip_pk", "departure_time"],
        unique=False,
    )
    op.create_table(
        "service_map_vertex",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("stop_pk", sa.Integer(), nullable=True),
        sa.Column("map_pk", sa.Integer(), nullable=True),
        sa.Column("position", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["map_pk"], ["service_map.pk"],),
        sa.ForeignKeyConstraint(["stop_pk"], ["stop.pk"],),
        sa.PrimaryKeyConstraint("pk"),
    )
    op.create_index(
        "service_map_vertex_map_pk_position",
        "service_map_vertex",
        ["map_pk", "position"],
        unique=False,
    )
    op.create_table(
        "trip_stop_time",
        sa.Column("pk", sa.Integer(), nullable=False),
        sa.Column("stop_pk", sa.Integer(), nullable=False),
        sa.Column("trip_pk", sa.Integer(), nullable=False),
        sa.Column("future", sa.Boolean(), nullable=True),
        sa.Column("arrival_time", sa.TIMESTAMP(timezone=True), nullable=True),
        sa.Column("arrival_delay", sa.Integer(), nullable=True),
        sa.Column("arrival_uncertainty", sa.Integer(), nullable=True),
        sa.Column("departure_time", sa.DateTime(timezone=True), nullable=True),
        sa.Column("departure_delay", sa.Integer(), nullable=True),
        sa.Column("departure_uncertainty", sa.Integer(), nullable=True),
        sa.Column("stop_sequence", sa.Integer(), nullable=False),
        sa.Column("track", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(["stop_pk"], ["stop.pk"],),
        sa.ForeignKeyConstraint(["trip_pk"], ["trip.pk"],),
        sa.PrimaryKeyConstraint("pk"),
        sa.UniqueConstraint("trip_pk", "stop_sequence"),
    )
    op.create_index(
        "trip_stop_time_stop_pk_arrival_time_idx",
        "trip_stop_time",
        ["stop_pk", "arrival_time"],
        unique=False,
    )


def downgrade():
    # op.drop_table("trip_stop_time")
    pass



import enum

import jinja2
import pytimeparse
import strictyaml
from strictyaml import (
    Bool,
    EmptyList,
    EmptyDict,
    Float,
    Map,
    MapPattern,
    Optional,
    ScalarValidator,
    Seq,
    Str,
)
from strictyaml.exceptions import YAMLSerializationError

from transiter import exceptions
from transiter import models
from transiter.services.servicemap import conditions


class HumanReadableTimePeriod(ScalarValidator):
    """
    A validator the converts human readable time period like "10 minutes" into the
    number of seconds in the period, as determined by pytimeparse.
    """

    def __init__(self):
        pass

    def validate_scalar(self, chunk):
        value = pytimeparse.parse(chunk.contents)
        if value is None:
            chunk.expecting_but_found(
                "when expecting something that could be interpreted as a time period",
                "found '{}'".format(chunk.contents),
            )
        return value

    @staticmethod
    def to_yaml(data):
        return "{} seconds".format(data)


class PyEnum(ScalarValidator):
    """
    A validator for enum.Enum types. This validator ensures that the provided string
    in the YAML file is the name of one of the enum's elements, and then casts the
    result to that enum element.
    """

    def __init__(self, enum_):
        self._enum = enum_
        assert issubclass(
            self._enum, enum.Enum
        ), "argument must be a enum.Enum or subclass thereof"

    def validate_scalar(self, chunk):
        try:
            val = self._enum[chunk.contents]
        except KeyError:
            chunk.expecting_but_found(
                "when expecting one of: {0}".format(
                    ", ".join(elem.name for elem in self._enum)
                )
            )
        else:
            return val

    def to_yaml(self, data):
        if not isinstance(data, self._enum):
            raise YAMLSerializationError(
                "Got '{0}' when  expecting one of: {1}".format(
                    data, ", ".join(str(elem) for elem in self._enum)
                )
            )
        return data.name

    def __repr__(self):
        return u"PyEnum({0})".format(repr(self._enum))


class ServiceMapConditions(Map):
    def __init__(self):
        super().__init__(
            {
                Optional(conditions.ALL_OF): self,
                Optional(conditions.ENDS_EARLIER_THAN): Float(),
                Optional(conditions.ENDS_LATER_THAN): Float(),
                Optional(conditions.NONE_OF): self,
                Optional(conditions.ONE_OF): self,
                Optional(conditions.STARTS_EARLIER_THAN): Float(),
                Optional(conditions.STARTS_LATER_THAN): Float(),
                Optional(conditions.WEEKDAY): Bool(),
                Optional(conditions.WEEKEND): Bool(),
            }
        )


# These are all constants so that reading the JSON response is less fragile
AUTO_UPDATE = "auto_update"
BUILT_IN = "built_in"
CONDITIONS = "conditions"
CUSTOM = "custom"
DIRECTION_RULES_FILES = "direction_rules_files"
ENABLED = "enabled"
FEEDS = "feeds"
HEADERS = "headers"
HTTP = "http"
NAME = "name"
PACKAGES = "packages"
PARSER = "parser"
PERIOD = "period"
REQUIRED_FOR_INSTALL = "required_for_install"
REQUIREMENTS = "requirements"
SERVICE_MAPS = "service_maps"
SETTINGS = "settings"
SOURCE = "source"
THRESHOLD = "threshold"
URL = "url"
USE_FOR_ROUTES_AT_STOP = "use_for_routes_at_stop"
USE_FOR_STOPS_IN_ROUTE = "use_for_stops_in_route"


default_service_map_config = {
    "all-times": {
        SOURCE: models.ServiceMapGroup.ServiceMapSource.SCHEDULE,
        THRESHOLD: 0.1,
        USE_FOR_STOPS_IN_ROUTE: True,
    },
    "weekday": {
        SOURCE: models.ServiceMapGroup.ServiceMapSource.SCHEDULE,
        THRESHOLD: 0.1,
        CONDITIONS: {conditions.WEEKDAY: True},
        USE_FOR_ROUTES_AT_STOP: True,
    },
    "weekend": {
        SOURCE: models.ServiceMapGroup.ServiceMapSource.SCHEDULE,
        THRESHOLD: 0.1,
        CONDITIONS: {conditions.WEEKEND: True},
        USE_FOR_ROUTES_AT_STOP: True,
    },
    "realtime": {
        SOURCE: models.ServiceMapGroup.ServiceMapSource.REALTIME,
        USE_FOR_STOPS_IN_ROUTE: True,
        USE_FOR_ROUTES_AT_STOP: True,
    },
}

_schema = Map(
    {
        NAME: Str(),
        Optional(REQUIREMENTS, {PACKAGES: [], SETTINGS: []}): Map(
            {
                Optional(PACKAGES, []): Seq(Str()) | EmptyList(),
                Optional(SETTINGS, []): Seq(Str()) | EmptyList(),
            }
        ),
        FEEDS: MapPattern(
            Str(),
            Map(
                {
                    HTTP: Map(
                        {
                            URL: Str(),
                            Optional(HEADERS, {}): MapPattern(Str(), Str())
                            | EmptyDict(),
                        }
                    ),
                    PARSER: Map(
                        {
                            Optional(BUILT_IN, None): PyEnum(models.Feed.BuiltInParser),
                            Optional(CUSTOM, None): Str(),
                        }
                    ),
                    Optional(AUTO_UPDATE, {ENABLED: False, PERIOD: -1}): Map(
                        {
                            Optional(ENABLED, True): Bool(),
                            Optional(PERIOD, -1): HumanReadableTimePeriod(),
                        }
                    ),
                    Optional(REQUIRED_FOR_INSTALL, False): Bool(),
                }
            ),
        ),
        Optional(SERVICE_MAPS, default_service_map_config): MapPattern(
            Str(),
            Map(
                {
                    SOURCE: PyEnum(models.ServiceMapGroup.ServiceMapSource),
                    Optional(THRESHOLD, 0): Float(),
                    Optional(CONDITIONS, None): ServiceMapConditions(),
                    Optional(USE_FOR_STOPS_IN_ROUTE, False): Bool(),
                    Optional(USE_FOR_ROUTES_AT_STOP, False): Bool(),
                }
            ),
        )
        | EmptyDict(),
    }
)


def render_template(jinja_yaml_template, setting_to_value=None):
    if setting_to_value is None:
        setting_to_value = {}
    try:
        return jinja2.Template(jinja_yaml_template).render(**setting_to_value)
    except jinja2.TemplateError as error:
        raise exceptions.InvalidSystemConfigFile(
            "The Jinja template is invalid\n" + str(error)
        )


def read(yaml_string, setting_to_value=None, label="transit system config"):
    if setting_to_value is None:
        setting_to_value = {}
    try:
        config = strictyaml.load(yaml_string, _schema, label=label).data
    except strictyaml.YAMLValidationError as error:
        raise exceptions.InvalidSystemConfigFile(
            "System config is valid YAML but is not a valid system config\n"
            + str(error)
        )
    except strictyaml.YAMLError as error:
        raise exceptions.InvalidSystemConfigFile(
            "Provided system config file cannot be parsed as YAML\n" + str(error)
        )

    required_settings = config[REQUIREMENTS][SETTINGS]
    missing_settings = set(required_settings) - set(setting_to_value.keys())
    if len(missing_settings) > 0:
        raise exceptions.InstallError(
            "Missing required settings: {}".format(", ".join(missing_settings))
        )

    return config



import enum

from sqlalchemy import Column, Enum, Float, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from .base import Base


class ServiceMapGroup(Base):
    __tablename__ = "service_map_group"

    pk = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    system_pk = Column(Integer, ForeignKey("system.pk"), nullable=False)

    class ServiceMapSource(enum.Enum):
        SCHEDULE = 1
        REALTIME = 2

    source = Column(Enum(ServiceMapSource, native_enum=False), nullable=False)
    conditions = Column(String)
    threshold = Column(Float, nullable=False, default=0)
    use_for_routes_at_stop = Column(Boolean, nullable=False, default=False)
    use_for_stops_in_route = Column(Boolean, nullable=False, default=False)

    maps = relationship(
        "ServiceMap", cascade="all, delete-orphan", back_populates="group"
    )
    system = relationship("System", cascade="", back_populates="service_map_groups")



import logging
import time
import typing

import sqlalchemy.sql.expression as sql
from sqlalchemy import func

from transiter import models
from transiter.data import dbconnection

logger = logging.getLogger(__name__)


def get_scheduled_trip_pk_to_path_in_system(system_pk):
    """
    Get a map of trip PK to the path of that trip for every scheduled trip in a system.
    By path is meant a list of stop_pks that the trip stops at. This method completely
    bypasses the ORM and so relatively efficient.

    :param system_pk: the system's PK
    :return: map of trip PK to list of stop PKs
    """
    start_time = time.time()
    session = dbconnection.get_session()
    query = (
        session.query(
            models.ScheduledTripStopTime.trip_pk, models.ScheduledTripStopTime.stop_pk
        )
        .join(
            models.ScheduledTrip,
            models.ScheduledTrip.pk == models.ScheduledTripStopTime.trip_pk,
        )
        .join(
            models.ScheduledService,
            sql.and_(
                models.ScheduledService.pk == models.ScheduledTrip.service_pk,
                models.ScheduledService.system_pk == system_pk,
            ),
        )
        .order_by(
            models.ScheduledTripStopTime.trip_pk,
            models.ScheduledTripStopTime.stop_sequence,
        )
    )
    trip_pk_to_stop_pks = {}
    for trip_pk, stop_pk in query:
        if trip_pk not in trip_pk_to_stop_pks:
            trip_pk_to_stop_pks[trip_pk] = []
        trip_pk_to_stop_pks[trip_pk].append(stop_pk)
    logger.info(
        "Query get_scheduled_trip_pk_to_path_in_system took {:.2} seconds".format(
            time.time() - start_time
        )
    )
    return trip_pk_to_stop_pks


def list_scheduled_trips_with_times_in_system(system_pk):
    """
    List scheduled trips in a system along with their start and end times.

    Start time is the minimum of the trips departure times; end time is the maximum.

    :param system_pk: the system's PK
    :return: list of three tuples (ScheduledTrip, start time, end time)
    """
    start_time = time.time()
    session = dbconnection.get_session()
    first_stop_query = (
        session.query(
            models.ScheduledTripStopTime.trip_pk.label("trip_pk"),
            func.min(models.ScheduledTripStopTime.departure_time).label("time"),
        )
        .group_by(models.ScheduledTripStopTime.trip_pk)
        .subquery()
    )
    last_stop_query = (
        session.query(
            models.ScheduledTripStopTime.trip_pk.label("trip_pk"),
            func.max(models.ScheduledTripStopTime.departure_time).label("time"),
        )
        .group_by(models.ScheduledTripStopTime.trip_pk)
        .subquery()
    )
    query = (
        session.query(
            models.ScheduledTrip, first_stop_query.c.time, last_stop_query.c.time
        )
        .join(
            models.ScheduledService,
            sql.and_(
                models.ScheduledService.pk == models.ScheduledTrip.service_pk,
                models.ScheduledService.system_pk == system_pk,
            ),
        )
        .join(first_stop_query, models.ScheduledTrip.pk == first_stop_query.c.trip_pk)
        .join(last_stop_query, models.ScheduledTrip.pk == last_stop_query.c.trip_pk)
    )
    result = query.all()
    logger.info(
        "Query list_scheduled_trips_with_times_in_system took {:.2} seconds".format(
            time.time() - start_time
        )
    )
    return result


def get_trip_id_to_pk_map_by_feed_pk(feed_pk):
    id_to_pk = {}
    query = (
        dbconnection.get_session()
        .query(models.ScheduledTrip.id, models.ScheduledTrip.pk)
        .filter(
            models.ScheduledService.pk == models.ScheduledTrip.service_pk,
            models.ScheduledService.source_pk == models.FeedUpdate.pk,
            models.FeedUpdate.feed_pk == feed_pk,
        )
    )
    for (id_, pk) in query.all():
        id_to_pk[id_] = pk
    return id_to_pk


def list_trips_by_system_pk_and_trip_ids(
    system_pk, trip_ids
) -> typing.List[models.ScheduledTrip]:
    """

    :param system_pk:
    :param trip_ids:
    :return: list of models.ScheduledTrip
    """

    query = (
        dbconnection.get_session()
        .query(models.ScheduledTrip)
        .filter(
            models.ScheduledTrip.id.in_(trip_ids),
            models.ScheduledService.pk == models.ScheduledTrip.service_pk,
            models.ScheduledService.system_pk == system_pk,
        )
    )
    return query.all()



"""
The GTFS Static Util contains the logic for reading feeds of this format.

The official reference is here: https://gtfs.org/reference/static
"""

import csv
import datetime
import enum
import io
import zipfile

from transiter import parse


# Additional arguments are accepted for forwards compatibility
# noinspection PyUnusedLocal
def parse_gtfs_static(binary_content, *args, **kwargs):
    """
    Parse a GTFS Static feed

    :param binary_content: raw binary GTFS static zip data
    """
    gtfs_static_file = GtfsStaticFile(binary_content)
    for parsing_function in [parse_routes, parse_stops, parse_schedule]:
        yield from parsing_function(gtfs_static_file)


class GtfsStaticFile:
    class _InternalFileName(enum.Enum):
        CALENDAR = "calendar.txt"
        ROUTES = "routes.txt"
        STOP_TIMES = "stop_times.txt"
        STOPS = "stops.txt"
        TRANSFERS = "transfers.txt"
        TRIPS = "trips.txt"

    def __init__(self, binary_content):
        self._zip_file = zipfile.ZipFile(io.BytesIO(binary_content))

    def calendar(self):
        return self._read_internal_file(self._InternalFileName.CALENDAR)

    def routes(self):
        return self._read_internal_file(self._InternalFileName.ROUTES)

    def stops(self):
        return self._read_internal_file(self._InternalFileName.STOPS)

    def stop_times(self):
        return self._read_internal_file(self._InternalFileName.STOP_TIMES)

    def transfers(self):
        return self._read_internal_file(self._InternalFileName.TRANSFERS)

    def trips(self):
        return self._read_internal_file(self._InternalFileName.TRIPS)

    def _read_internal_file(self, file_name):
        """
        Read a GTFS static file

        :param file_name: which static file to real
        :return: iterator of dictionaries for each row
        """
        file_name = file_name.value
        try:
            with self._zip_file.open(file_name) as raw_csv_file:
                csv_file = io.TextIOWrapper(raw_csv_file, "utf-8-sig")
                csv_reader = csv.DictReader(csv_file)
                for row in csv_reader:
                    yield row
        except KeyError:
            return []


def parse_routes(gtfs_static_file: GtfsStaticFile):
    for row in gtfs_static_file.routes():
        yield parse.Route(
            id=row["route_id"],
            type=parse.Route.Type(int(row["route_type"])),
            color=row.get("route_color", "FFFFF"),
            text_color=row.get("route_text_color", "000000"),
            url=row.get("route_url"),
            short_name=row.get("route_short_name"),
            long_name=row.get("route_long_name"),
            description=row.get("route_desc"),
            sort_order=row.get("route_sort_order"),
        )


def parse_stops(gtfs_static_file: GtfsStaticFile):

    stop_id_to_stop = {}
    stop_id_to_parent_stop_id = {}

    # Step 1: read the basic stops in the GTFS feed into Stop objects.
    for row in gtfs_static_file.stops():
        stop = parse.Stop(
            id=row["stop_id"],
            name=row["stop_name"],
            longitude=float(row["stop_lon"]),
            latitude=float(row["stop_lat"]),
            url=row.get("stop_url"),
            is_station=row.get("location_type") == "1",
        )
        parent_stop_id = row.get("parent_station", "")
        if parent_stop_id != "":
            stop_id_to_parent_stop_id[stop.id] = parent_stop_id
        stop_id_to_stop[stop.id] = stop

    for stop_id, parent_stop_id in stop_id_to_parent_stop_id.items():
        stop_id_to_stop[stop_id].parent_stop = stop_id_to_stop[parent_stop_id]

    # Step 2: replace the parent stop IDs with the actual parent stop. If a stop does
    # not have a parent, make it a station.
    stop_id_to_station_id = {}
    station_sets_by_stop_id = {}
    for stop in stop_id_to_stop.values():
        if stop.parent_stop is None:
            stop.is_station = True
        if stop.is_station:
            station_sets_by_stop_id[stop.id] = {stop.id}
        else:
            stop_id_to_station_id[stop.id] = stop.parent_stop.id
        yield stop

    # Step 3: using the GTFS transfers data, link together stops which have a free
    # transfer.
    for row in gtfs_static_file.transfers():
        stop_id_1 = row["from_stop_id"]
        stop_id_2 = row["to_stop_id"]
        if stop_id_1 == stop_id_2:
            continue
        updated_station_set = station_sets_by_stop_id[stop_id_1].union(
            station_sets_by_stop_id[stop_id_2]
        )
        for stop_id in updated_station_set:
            station_sets_by_stop_id[stop_id] = updated_station_set

    # Step 4: create parent stations for stop linked together in Step 3.
    for station_set in station_sets_by_stop_id.values():
        if len(station_set) <= 1:
            continue
        child_stops = [stop_id_to_stop[stop_id] for stop_id in station_set]
        parent_stop = create_station_from_child_stops(child_stops)
        for child_stop in child_stops:
            child_stop.parent_stop = parent_stop
        yield parent_stop
        station_set.clear()


def create_station_from_child_stops(child_stops):
    """
    Create a station from child stops.

    Some attributes will be set automatically based on the children:
    - latitude: the average of the child stops' latitudes
    - longitude: same
    - id: A-B-C, where [A, B, C] are the child stops' ids in sorted order.
    - name: grouping children by name, if there is unique largest group then
        this is set to be the parents name. Otherwise its the names in the
        largest groups in sorted order separated by forward slash.

    :param child_stops: list of child stops
    :return: the parent station
    """
    latitude = sum(float(child_stop.latitude) for child_stop in child_stops) / len(
        child_stops
    )
    longitude = sum(float(child_stop.longitude) for child_stop in child_stops) / len(
        child_stops
    )

    child_stop_ids = [child_stop.id for child_stop in child_stops]
    stop_id = "-".join(sorted(child_stop_ids))

    child_stop_names = {child_stop.name: 0 for child_stop in child_stops}
    for child_stop in child_stops:
        child_stop_names[child_stop.name] += 1
    max_freq = max(child_stop_names.values())
    most_frequent_names = set()
    for child_stop_name, freq in child_stop_names.items():
        if freq == max_freq:
            most_frequent_names.add(child_stop_name)

    # This part removes names which appear as substrings in other names
    for name in most_frequent_names.copy():
        remove = False
        for other_name in most_frequent_names:
            if name != other_name and name in other_name:
                remove = True
        if remove:
            most_frequent_names.remove(name)
    name = " / ".join(sorted(most_frequent_names))

    return parse.Stop(
        id=stop_id, name=name, longitude=longitude, latitude=latitude, is_station=True
    )


def parse_schedule(gtfs_static_file: GtfsStaticFile):
    str_to_bool = {"0": False, "1": True}

    service_id_to_service = {}
    for row in gtfs_static_file.calendar():
        service_id = row["service_id"]
        service_id_to_service[service_id] = parse.ScheduledService(
            id=service_id,
            monday=str_to_bool[row["monday"]],
            tuesday=str_to_bool[row["tuesday"]],
            wednesday=str_to_bool[row["wednesday"]],
            thursday=str_to_bool[row["thursday"]],
            friday=str_to_bool[row["friday"]],
            saturday=str_to_bool[row["saturday"]],
            sunday=str_to_bool[row["sunday"]],
        )

    trip_id_to_trip = {}
    for row in gtfs_static_file.trips():
        service_id = row["service_id"]
        if service_id not in service_id_to_service:
            continue
        trip = parse.ScheduledTrip(
            id=row["trip_id"],
            route_id=row["route_id"],
            direction_id=str_to_bool.get(row["direction_id"]),
        )
        service_id_to_service[service_id].trips.append(trip)
        trip_id_to_trip[trip.id] = trip

    # NOTE: memoization of the next function cuts about 2 seconds off the time taken to
    # parse the NYC Subway's GTFS static feed. However because the function itself is
    # not very computationally intensive, to see any benefit it is necessary to have
    # a very simple memoization process.
    cache = {}

    def time_string_to_datetime_time(time_string):
        if time_string not in cache:
            hour, minute, second = time_string.split(":")
            cache[time_string] = datetime.time(
                hour=int(hour) % 24, minute=int(minute), second=int(second)
            )
        return cache[time_string]

    for row in gtfs_static_file.stop_times():
        trip_id = row["trip_id"]
        if trip_id not in trip_id_to_trip:
            continue
        stop_time = parse.ScheduledTripStopTime(
            stop_id=row["stop_id"],
            stop_sequence=int(row["stop_sequence"]),
            departure_time=time_string_to_datetime_time(row["departure_time"]),
            arrival_time=time_string_to_datetime_time(row["arrival_time"]),
        )
        trip_id_to_trip[trip_id].stop_times.append(stop_time)

    yield from service_id_to_service.values()



from transiter.parse import gtfsrealtime, utils

create_parser = gtfsrealtime.create_parser
built_in_parser = gtfsrealtime.built_in_parser
read_gtfs_realtime = gtfsrealtime.read_gtfs_realtime
transform_to_transiter_structure = gtfsrealtime.transform_to_transiter_structure
TripDataCleaner = utils.TripDataCleaner



from typing import Dict, Iterable

from sqlalchemy import sql
from sqlalchemy.orm import joinedload, selectinload

from transiter import models
from transiter.data import dbconnection
from transiter.data.dams import genericqueries


def list_all_in_system(system_id):
    """
    List all stops in a system.

    :param system_id: the system's ID
    :return: a list of Stops
    """
    return genericqueries.list_all_in_system(models.Stop, system_id, models.Stop.id)


def get_in_system_by_id(system_id, stop_id):
    """
    Get a specific stop in a system.

    :param system_id: the system's ID
    :param stop_id: the stop's ID
    :return: Stop, if it exists; None if it does not
    """
    return genericqueries.get_in_system_by_id(models.Stop, system_id, stop_id)


def get_id_to_pk_map_in_system(
    system_pk: int, stop_ids: Iterable[str] = None
) -> Dict[str, int]:
    """
    Get a map of stop ID to stop PK for all stops in a system.
    """
    return genericqueries.get_id_to_pk_map(models.Stop, system_pk, stop_ids)


def list_direction_rules_for_stops(stop_pks):
    """
    List all direction rules for a given collection of stops.
    """
    return (
        dbconnection.get_session()
        .query(models.DirectionRule)
        .filter(models.DirectionRule.stop_pk.in_(stop_pks))
        .all()
    )


def list_all_stops_in_stop_tree(stop_pk) -> Iterable[models.Stop]:
    """
    List all stops in the stop tree of a given stop.
    """
    session = dbconnection.get_session()

    # The first CTE retrieves all stop PKs of ancestors of the root stop.
    ancestor_cte = (
        session.query(models.Stop.pk, models.Stop.parent_stop_pk)
        .filter(models.Stop.pk == stop_pk)
        .cte(name="ancestor", recursive=True)
    )
    ancestor_cte = ancestor_cte.union_all(
        session.query(models.Stop.pk, models.Stop.parent_stop_pk).filter(
            models.Stop.pk == ancestor_cte.c.parent_stop_pk
        )
    )

    # The second CTE then retrieves all descendents of all stops from the first CTE.
    # Because the first CTE returns the root of the stops tree, the second CTE returns
    # all stops in the tree.
    relation_cte = (
        session.query(models.Stop.pk, models.Stop.parent_stop_pk)
        .filter(models.Stop.pk == ancestor_cte.c.pk)
        .cte(name="relation", recursive=True)
    )
    relation_cte = relation_cte.union_all(
        session.query(models.Stop.pk, models.Stop.parent_stop_pk).filter(
            models.Stop.parent_stop_pk == relation_cte.c.pk
        )
    )
    query = session.query(models.Stop).filter(models.Stop.pk == relation_cte.c.pk)
    return query.all()


def list_stop_time_updates_at_stops(stop_pks, earliest_time=None, latest_time=None):
    """
    List the future TripStopTimes for a collection of stops.

    The list is ordered by departure time and, in the case of ties, by
    the arrival time.

    Note the Trip and Route objects are eagerly loaded.
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.TripStopTime)
        .filter(models.TripStopTime.stop_pk.in_(stop_pks))
        .filter(models.TripStopTime.trip_pk == models.Trip.pk)
        .filter(models.TripStopTime.future)
        .order_by(models.TripStopTime.departure_time)
        .order_by(models.TripStopTime.arrival_time)
        .options(joinedload(models.TripStopTime.trip))
        .options(selectinload(models.TripStopTime.trip, models.Trip.route))
    )

    if earliest_time is not None:
        earliest_datetime = earliest_time
        query = query.filter(
            sql.or_(
                models.TripStopTime.departure_time >= earliest_datetime,
                models.TripStopTime.arrival_time >= earliest_datetime,
            )
        )
    if latest_time is not None:
        latest_datetime = latest_time
        query = query.filter(
            sql.or_(
                models.TripStopTime.departure_time <= latest_datetime,
                models.TripStopTime.arrival_time <= latest_datetime,
            )
        )
    return query.all()


def get_stop_pk_to_station_pk_map_in_system(system_id):
    """
    Get the map of stop PK to station PK for every stop in a system.

    Right now this method assumes that a stop's station is either itself
    or its parent.

    :param system_id: the system ID
    :return: map of stop PK to stop PK
    """
    session = dbconnection.get_session()
    query = (
        session.query(
            models.Stop.pk, models.Stop.parent_stop_pk, models.Stop.is_station
        )
        .filter(models.Stop.system_pk == models.System.pk)
        .filter(models.System.id == system_id)
    )
    stop_pk_to_station_pk = {}
    for stop_pk, parent_stop_pk, is_station in query:
        if is_station or parent_stop_pk is None:
            stop_pk_to_station_pk[stop_pk] = stop_pk
        else:
            stop_pk_to_station_pk[stop_pk] = parent_stop_pk
    return stop_pk_to_station_pk



from sqlalchemy import (
    Column,
    Enum,
    TIMESTAMP,
    Integer,
    String,
    Boolean,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.Trip)
class Trip(Base):
    __tablename__ = "trip"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    route_pk = Column(Integer, ForeignKey("route.pk"), nullable=False)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    TripStatus = parse.Trip.Status

    direction_id = Column(Boolean)
    start_time = Column(TIMESTAMP(timezone=True))

    last_update_time = Column(TIMESTAMP(timezone=True))
    vehicle_id = Column(String)
    current_status = Column(Enum(TripStatus, name="tripstatus", native_enum=False))
    current_stop_sequence = Column(Integer)

    source = relationship("FeedUpdate", cascade="none")
    route = relationship("Route", back_populates="trips", cascade="none")
    stop_times = relationship(
        "TripStopTime",
        back_populates="trip",
        order_by="TripStopTime.stop_sequence",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
    )

    __table_args__ = (UniqueConstraint(route_pk, id),)



from typing import Optional

from sqlalchemy import sql
from sqlalchemy.orm import joinedload

from transiter import models
from transiter.data import dbconnection
from transiter.data.dams import genericqueries


def list_all_feed_pks():
    return list(pk_ for (pk_,) in dbconnection.get_session().query(models.Feed.pk))


def list_all_auto_updating():
    """
    List all auto-updating Feeds.

    :return: list of Feeds
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.Feed)
        .join(models.System, models.System.pk == models.Feed.system_pk)
        .filter(models.Feed.auto_update_enabled)
        .filter(models.System.auto_update_enabled)
        .filter(models.System.status == models.System.SystemStatus.ACTIVE)
    )
    return query.all()


def list_all_in_system(system_id):
    """
    List all Feeds in a System.

    :param system_id: the system's ID
    :return: list of Feeds
    """
    return genericqueries.list_all_in_system(models.Feed, system_id, models.Feed.id)


def get_in_system_by_id(system_id, feed_id):
    """
    Get a Feed in a System.

    :param system_id: the system's ID
    :param feed_id: the feed's ID
    :return: Feed, if it exists; None, otherwise
    """
    return genericqueries.get_in_system_by_id(models.Feed, system_id, feed_id)


def get_update_by_pk(feed_update_pk) -> Optional[models.FeedUpdate]:
    session = dbconnection.get_session()
    return (
        session.query(models.FeedUpdate)
        .filter(models.FeedUpdate.pk == feed_update_pk)
        .options(joinedload(models.FeedUpdate.feed))
        .options(joinedload(models.FeedUpdate.feed, models.Feed.system))
        .one_or_none()
    )


def get_last_successful_update_hash(feed_pk) -> Optional[str]:
    """
    Get the last successful FeedUpdate content hash for a Feed.

    :param feed_pk: the feed's PK
    :return: FeedUpdate, or None if it doesn't exist
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.FeedUpdate.content_hash)
        .filter(models.FeedUpdate.feed_pk == feed_pk)
        .order_by(models.FeedUpdate.completed_at.desc())
        .filter(models.FeedUpdate.status == "SUCCESS")
        .limit(1)
    )
    result = query.first()
    if result is None:
        return None
    return result[0]


def list_updates_in_feed(feed_pk):
    """
    List the most recent updates in a feed, ordered descending in time.

    :param feed_pk: the Feed's PK
    :return: list of FeedUpdates
    """
    session = dbconnection.get_session()
    query = (
        session.query(models.FeedUpdate)
        .filter(models.FeedUpdate.feed_pk == feed_pk)
        .order_by(models.FeedUpdate.pk.desc())
        .limit(100)
    )
    return query.all()


def trim_feed_updates(feed_pk, before_datetime):
    """
    Trip all FeedUpdates for a feed whose last action time was before
    a certain cut-off point.

    :param feed_pk: pk of the feed
    :param before_datetime: the cut-off point
    """
    not_exists_conditions = [
        ~sql.exists(
            sql.select([sql.literal_column("1")]).where(
                UpdatableEntity.source_pk == models.FeedUpdate.pk
            )
        )
        for UpdatableEntity in models.list_updatable_entities()
    ]
    query = sql.delete(models.FeedUpdate).where(
        sql.and_(
            models.FeedUpdate.feed_pk == feed_pk,
            models.FeedUpdate.completed_at <= before_datetime,
            *not_exists_conditions
        )
    )
    dbconnection.get_session().execute(query)


def delete_in_system_by_id(system_id, feed_id):
    """
    Delete a Feed from the DB whose ID is given.

    :return: True if an entity was found and deleted, false if no such
     entity exists
    """
    session = dbconnection.get_session()
    entity = get_in_system_by_id(system_id, feed_id)
    if entity is None:
        return False
    session.delete(entity)
    return True



import typing

from transiter import exceptions, models
from transiter.data import dbconnection
from transiter.data.dams import tripdam, routedam
from transiter.services import views


@dbconnection.unit_of_work
def list_all_in_route(system_id, route_id) -> typing.List[views.Trip]:
    """
    Get representations for all trips in a system.
    """
    route = routedam.get_in_system_by_id(system_id, route_id)
    if route is None:
        raise exceptions.IdNotFoundError(
            models.Route, system_id=system_id, route_id=route_id
        )

    response = []
    trips = tripdam.list_all_in_route_by_pk(route.pk)
    trip_pk_to_last_stop = tripdam.get_trip_pk_to_last_stop_map(
        trip.pk for trip in trips
    )
    for trip in trips:
        trip_response = views.Trip.from_model(trip)
        last_stop = trip_pk_to_last_stop.get(trip.pk)
        if last_stop is not None:
            trip_response.last_stop = views.Stop.from_model(last_stop)
        response.append(trip_response)
    return response


@dbconnection.unit_of_work
def get_in_route_by_id(system_id, route_id, trip_id):
    """
    Get a representation for a trip in a system
    """
    trip = tripdam.get_in_route_by_id(system_id, route_id, trip_id)
    if trip is None:
        raise exceptions.IdNotFoundError(
            models.Trip, system_id=system_id, route_id=route_id, trip_id=trip_id
        )
    trip_response = views.Trip.from_model(trip)
    trip_response.route = views.Route.from_model(trip.route)
    trip_response.stop_times = []
    for stop_time in trip.stop_times:
        stop_time_response = views.TripStopTime.from_model(stop_time)
        stop_time_response.stop = views.Stop.from_model(stop_time.stop)
        trip_response.stop_times.append(stop_time_response)
    return trip_response



"""
Transiter Scheduler

The scheduler is a Python process that runs tasks periodically using APScheduler. It
has a HTTP interface that enables a Transiter webserver to communicate with it.
"""
import collections
import datetime
import json
import logging
import random
import time
import typing

import apscheduler.schedulers.background
import flask
import inflection
from sqlalchemy.exc import SQLAlchemyError

from transiter.executor import celeryapp
from transiter.services import feedservice

logger = logging.getLogger("transiter")


scheduler = apscheduler.schedulers.background.BackgroundScheduler(
    executors={"default": {"type": "threadpool", "max_workers": 20}}
)


class Schedule:
    def job_kwargs(self):
        raise NotImplementedError

    def json(self):
        return {
            "type": inflection.underscore(self.__class__.__name__)[:-9].upper(),
            "parameters": {key[1:]: value for key, value in vars(self).items()},
        }

    def __repr__(self):
        return ",".join(
            sorted("{}={}".format(*item) for item in self.job_kwargs().items())
        )

    def __eq__(self, other):
        return isinstance(other, Schedule) and str(self) == str(other)


class PeriodicSchedule(Schedule):
    def __init__(self, period):
        self._period = period

    @property
    def period(self):
        return self._period

    def job_kwargs(self):
        return {
            "trigger": "interval",
            "seconds": self._period,
            "next_run_time": self._calculate_next_run_time(self._period),
        }

    @staticmethod
    def _calculate_next_run_time(period):
        return datetime.datetime.now() + datetime.timedelta(
            seconds=period * random.uniform(0, 1)
        )


class CronSchedule(Schedule):
    def __init__(self, minute):
        self._minute = minute

    def job_kwargs(self):
        return {"trigger": "cron", "minute": self._minute}


class Task:
    def __init__(self):
        self._schedule: typing.Optional[Schedule] = None
        self._job = None

    def run(self):
        raise NotImplementedError

    @property
    def schedule(self):
        return self._schedule

    @schedule.setter
    def schedule(self, schedule: typing.Optional[Schedule]):
        if self._schedule == schedule:
            return
        if self._schedule is not None:
            self._job.remove()
        self._schedule = schedule
        if schedule is None:
            return
        self._job = scheduler.add_job(self.run, **schedule.job_kwargs())

    def json(self):
        return {
            "type": inflection.underscore(self.__class__.__name__)[:-5].upper(),
            "parameters": {
                key[1:]: value
                for key, value in vars(self).items()
                if key not in {"_job", "_schedule"}
            },
            "schedule": self._schedule.json() if self._schedule is not None else None,
        }


class FeedAutoUpdateTask(Task):
    def __init__(self, system_id, feed_id):
        self._system_id = system_id
        self._feed_id = feed_id
        super().__init__()

    def run(self):
        expires = (
            self.schedule.period * 0.8
            if isinstance(self.schedule, PeriodicSchedule)
            else None
        )
        # NOTE: access the task like a static method to avoid the current instance
        # being unnecessarily being sent through RabbitMQ.
        FeedAutoUpdateTask._create_feed_update.apply_async(
            args=(None, self._system_id, self._feed_id), expires=expires
        )

    @celeryapp.app.task
    def _create_feed_update(self, system_id, feed_id):
        return feedservice.create_and_execute_feed_update(
            system_id, feed_id, execute_async=False
        )


class TrimFeedUpdatesTask(Task):
    def __init__(self):
        super().__init__()
        self.schedule = CronSchedule(minute="*/15")

    def run(self):
        # NOTE: access the task like a static method to avoid the current instance
        # being unnecessarily being sent through RabbitMQ.
        TrimFeedUpdatesTask._trim_feed_updates.delay(None)

    @celeryapp.app.task
    def _trim_feed_updates(self):
        return feedservice.trim_feed_updates()


class Registry:
    def initialize(self):
        while True:
            try:
                self.refresh()
            except SQLAlchemyError:
                logger.info(
                    "Failed to connect to the database; trying again in 1 second."
                )
                time.sleep(1)
                continue
            except Exception as e:
                logger.exception("Unexpected error", e)
                exit(1)
            break

    def refresh(self):
        raise NotImplementedError

    def all_tasks(self):
        raise NotImplementedError


class FeedAutoUpdateRegistry(Registry):
    def __init__(self):
        self._system_id_to_feed_id_to_task: typing.Dict[
            str, typing.Dict[str, FeedAutoUpdateTask]
        ] = collections.defaultdict(lambda: {})

    def refresh(self):
        stale_feed_keys = set(self._current_feed_keys())
        for feed in feedservice.list_all_auto_updating():
            task = self._system_id_to_feed_id_to_task[feed.system.id].get(feed.id)
            if task is None:
                logger.info("Adding {}/{}".format(feed.system.id, feed.id))
                task = FeedAutoUpdateTask(feed.system.id, feed.id)
                self._system_id_to_feed_id_to_task[feed.system.id][feed.id] = task
            else:
                logger.info("Updating {}/{}".format(feed.system.id, feed.id))
            task.schedule = PeriodicSchedule(feed.auto_update_period)
            stale_feed_keys.discard((feed.system.id, feed.id))

        for system_id, feed_id in stale_feed_keys:
            logger.info("Removing {}/{}".format(system_id, feed_id))
            feed_id_to_task = self._system_id_to_feed_id_to_task[system_id]
            feed_id_to_task[feed_id].schedule = None
            feed_id_to_task.pop(feed_id)
            if len(feed_id_to_task) == 0:
                del self._system_id_to_feed_id_to_task[system_id]

    def all_tasks(self):
        for feed_id_to_task in self._system_id_to_feed_id_to_task.values():
            yield from feed_id_to_task.values()

    def _current_feed_keys(self):
        for system_id, feed_id_to_task in self._system_id_to_feed_id_to_task.items():
            for feed_id in feed_id_to_task.keys():
                yield system_id, feed_id


class TransiterRegistry(Registry):
    def __init__(self):
        self._tasks = []

    def refresh(self):
        if len(self._tasks) > 0:
            return
        self._tasks = [TrimFeedUpdatesTask()]

    def all_tasks(self):
        return self._tasks


feed_auto_update_registry = FeedAutoUpdateRegistry()
transiter_registry = TransiterRegistry()


def app_ping():
    logger.info("Received external ping via HTTP")
    return json.dumps(
        [
            task.json()
            for registry in [feed_auto_update_registry, transiter_registry]
            for task in registry.all_tasks()
        ],
        indent=2,
    )


def app_refresh_tasks():
    logger.info("Received external refresh tasks command via HTTP")
    feed_auto_update_registry.refresh()
    return "", 204


def create_app():
    app = flask.Flask(__name__)
    app.add_url_rule("/", "ping", app_ping, methods=["GET"])
    app.add_url_rule("/", "refresh_tasks", app_refresh_tasks, methods=["POST"])

    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    logger.addHandler(handler)
    formatter = logging.Formatter(
        "%(asctime)s TS %(levelname)-5s [%(module)s] %(message)s"
    )
    handler.setFormatter(formatter)

    logger.info("Launching scheduler")
    scheduler.start()
    feed_auto_update_registry.initialize()
    transiter_registry.initialize()

    logger.info("Launching HTTP server")
    return app



"""
The scheduler client module is used to talk with the scheduler.
"""
import json
import logging

import requests

from transiter import config

logger = logging.getLogger(__name__)


def ping():
    """
    Ping the scheduler.

    If it cannot be reached return None; otherwise, return a JSON representation of the
    tasks that are currently being scheduled.
    """
    try:
        response = requests.get(
            "http://{}:{}".format(config.SCHEDULER_HOST, config.SCHEDULER_PORT),
            timeout=0.25,
        )
        response.raise_for_status()
        return json.loads(response.text)
    except requests.RequestException:
        return None


def refresh_tasks():
    """
    Refresh the scheduler's update task list.
    """
    try:
        response = requests.post(
            "http://{}:{}".format(config.SCHEDULER_HOST, config.SCHEDULER_PORT)
        )
        response.raise_for_status()
        return True
    except requests.RequestException:
        logger.info("Could not connect to the Transiter scheduler")
    return False



from sqlalchemy import (
    Column,
    TIMESTAMP,
    DateTime,
    Index,
    Integer,
    String,
    Boolean,
    UniqueConstraint,
    ForeignKey,
)
from sqlalchemy.orm import relationship

from .base import Base


class TripStopTime(Base):
    __tablename__ = "trip_stop_time"

    pk = Column(Integer, primary_key=True)
    stop_pk = Column(Integer, ForeignKey("stop.pk"), nullable=False)
    trip_pk = Column(Integer, ForeignKey("trip.pk"), nullable=False)

    future = Column(Boolean, default=True, nullable=False)
    arrival_time = Column(TIMESTAMP(timezone=True))
    arrival_delay = Column(Integer)
    arrival_uncertainty = Column(Integer)
    departure_time = Column(DateTime(timezone=True))
    departure_delay = Column(Integer)
    departure_uncertainty = Column(Integer)
    stop_sequence = Column(Integer, nullable=False)
    track = Column(String)

    stop_id = None

    stop = relationship("Stop", back_populates="trip_times", cascade="none")
    trip = relationship(
        "Trip", back_populates="stop_times", cascade="none", cascade_backrefs=False
    )

    __table_args__ = (
        UniqueConstraint(trip_pk, stop_sequence),
        Index("trip_stop_time_stop_pk_arrival_time_idx", stop_pk, arrival_time),
    )

    def get_time(self):
        return self.arrival_time or self.departure_time



import flask

from transiter.http.httpmanager import http_endpoint, link_target, get_url_parameters
from transiter.services import stopservice, views

stop_endpoints = flask.Blueprint(__name__, __name__)


@http_endpoint(stop_endpoints, "")
@link_target(views.StopsInSystem, ["_system_id"])
def list_all_in_system(system_id):
    """List all stops for a specific system."""
    return stopservice.list_all_in_system(system_id)


@http_endpoint(stop_endpoints, "/<stop_id>")
@link_target(views.Stop, ["_system_id", "id"])
def get_in_system_by_id(system_id, stop_id):
    """Retrieve a specific stop in a specific system."""
    request_args = get_url_parameters(
        ["minimum_number_of_trips", "include_all_trips_within", "exclude_trips_before"]
    )
    return stopservice.get_in_system_by_id(system_id, stop_id, **request_args)



"""
Add a non-null constraint to trip_stop_time.

Revision ID: be93cf805bb4
Revises: 14fa42915420
Create Date: 2020-03-14 19:26:57.258062

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "be93cf805bb4"
down_revision = "14fa42915420"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        "trip_stop_time",
        "future",
        existing_type=sa.BOOLEAN(),
        nullable=False,
        # The server default is just set for the migration to update existing
        # records and not fail on adding the constraint.
        server_default="True",
    )
    op.alter_column("trip_stop_time", "future", server_default=None)


def downgrade():
    pass



"""
The sync module is responsible for syncing the results of feed parsers to the database.

This involves:

- Persisting entities in the feed that don't correspond to existing entities.

- Updating existing entities with new data in the feed. In certain cases, for example
    Trips, old data such as past arrival times is preserved.

- Deleting existing entities that no longer appear in the feed.
"""
import collections
import dataclasses
import logging
import typing
from typing import Iterable, List, Tuple

from transiter import models, parse
from transiter.data import dbconnection
from transiter.data.dams import genericqueries
from transiter.data.dams import stopdam, tripdam, routedam, scheduledam, feeddam
from transiter.models import updatableentity
from transiter.services.servicemap import servicemapmanager
from transiter.services.update import fastscheduleoperations

logger = logging.getLogger(__name__)


def sync(feed_update_pk, parser_object: parse.TransiterParser):
    """
    Sync entities to the database.

    :param feed_update_pk: the feed update event in which this sync operation is being
      performed
    :param parser_object: the parser object
    """
    feed_update = feeddam.get_update_by_pk(feed_update_pk)

    syncers_in_order = [
        RouteSyncer,
        StopSyncer,
        ScheduleSyncer,
        DirectionRuleSyncer,
        TripSyncer,
        AlertSyncer,
    ]
    if feed_update.update_type == feed_update.Type.FLUSH:
        syncers_in_order.reverse()

    totals = [0, 0, 0]
    for syncer_class in syncers_in_order:
        if syncer_class.feed_entity() not in parser_object.supported_types:
            continue
        logger.debug("Syncing {}".format(syncer_class.feed_entity()))
        entities = list(parser_object.get_entities(syncer_class.feed_entity()))
        for entity in entities:
            entity.source_pk = feed_update.pk
        result = syncer_class(feed_update).run(entities)
        for i in range(3):
            totals[i] += result[i]
    return tuple(totals)


class SyncerBase:

    __feed_entity__ = None
    __db_entity__ = None

    def __init__(self, feed_update: models.FeedUpdate):
        self.feed_update = feed_update

    def run(self, entities):
        self.pre_sync()
        if len(entities) > 0:
            num_added, num_updated = self.sync(entities)
        else:
            num_added, num_updated = 0, 0
        num_deleted = self.delete_stale_entities()
        self.post_sync()
        return num_added, num_updated, num_deleted

    def pre_sync(self):
        pass

    def sync(self, entities):
        raise NotImplementedError  # pragma: no cover

    def delete_stale_entities(self):
        """
        Remove stale entities from the database.

        A stale entity is currently defined as an entity that was most recently added or
        updated by the feed associated to the current feed update, but that was not
        contained in the current feed update. I.e., it is an entity which previously
        appeared in the current feed of interest but has since disappeared.
        """
        assert self.__db_entity__ is not None
        session = dbconnection.get_session()
        entities_to_delete = genericqueries.list_stale_entities(
            self.__db_entity__, self.feed_update
        )
        for entity in entities_to_delete:
            session.delete(entity)
        return len(entities_to_delete)

    def post_sync(self):
        pass

    @classmethod
    def feed_entity(cls):
        return cls.__feed_entity__

    def _merge_entities(self, entities) -> Tuple[list, int, int]:
        """
        Merge entities of a given type into the session.

        This function will merge entities such that, for example, an existing Alert in
        the feed will not be added but instead its preexisting DB version will be updated.

        :param entities: the entities to merge
        :return: the persisted entities
        """

        persisted_entities = []
        id_to_pk = self._get_id_to_pk_map()
        session = dbconnection.get_session()
        num_updated_entities = 0
        for entity in entities:
            if entity.id in id_to_pk:
                num_updated_entities += 1
                entity.pk = id_to_pk[entity.id]
            entity.source_pk = self.feed_update.pk
            persisted_entities.append(session.merge(entity))
        return (
            persisted_entities,
            len(entities) - num_updated_entities,
            num_updated_entities,
        )

    def _get_id_to_pk_map(self):
        return genericqueries.get_id_to_pk_map(
            self.__db_entity__, self.feed_update.feed.system.pk
        )


def syncer(entity_type) -> typing.Type[SyncerBase]:
    class _Syncer(SyncerBase):
        __db_entity__ = entity_type
        __feed_entity__ = updatableentity.get_feed_entity(entity_type)

        def sync(self, entities):
            raise NotImplementedError

    return _Syncer


class RouteSyncer(syncer(models.Route)):
    def sync(self, parsed_routes):
        routes = list(map(models.Route.from_parsed_route, parsed_routes))
        for route in routes:
            route.system_pk = self.feed_update.feed.system_pk
        __, num_added, num_updated = self._merge_entities(routes)
        return num_added, num_updated


class StopSyncer(syncer(models.Stop)):
    def sync(self, parsed_stops: typing.Iterable[parse.Stop]):
        # NOTE: the stop tree is manually linked together because otherwise SQL
        # Alchemy's cascades will result in duplicate entries in the DB because the
        # models do not have PKs yet.
        stop_id_to_parent_stop_id = {}
        stops = []
        for parsed_stop in parsed_stops:
            stop = models.Stop.from_parsed_stop(parsed_stop)
            stop.system_pk = self.feed_update.feed.system_pk
            if parsed_stop.parent_stop is not None:
                stop_id_to_parent_stop_id[parsed_stop.id] = parsed_stop.parent_stop.id
            else:
                stop_id_to_parent_stop_id[parsed_stop.id] = None
            stops.append(stop)
        persisted_stops, num_added, num_updated = self._merge_entities(stops)
        stop_id_to_persisted_stops = {stop.id: stop for stop in persisted_stops}

        # NOTE: flush the session the session to populate the primary keys
        dbconnection.get_session().flush()
        for stop_id in stop_id_to_parent_stop_id.keys():
            stop = stop_id_to_persisted_stops[stop_id]
            parent_stop = stop_id_to_persisted_stops.get(
                stop_id_to_parent_stop_id.get(stop.id)
            )
            if parent_stop is not None:
                stop.parent_stop_pk = parent_stop.pk
            else:
                stop.parent_stop_pk = None
        return num_added, num_updated


class ScheduleSyncer(syncer(models.ScheduledService)):

    recalculate_service_maps = False

    def pre_sync(self):
        num_entities_deleted = fastscheduleoperations.delete_trips_associated_to_feed(
            self.feed_update.feed.pk
        )
        if num_entities_deleted > 0:
            self.recalculate_service_maps = True

    def sync(self, parsed_services):
        persisted_services, num_added, num_updated = self._merge_entities(
            list(map(models.ScheduledService.from_parsed_service, parsed_services))
        )
        for service in persisted_services:
            service.system = self.feed_update.feed.system
        schedule_updated = fastscheduleoperations.sync_trips(
            self.feed_update, parsed_services
        )
        if schedule_updated:
            self.recalculate_service_maps = True
        return num_added, num_updated

    def post_sync(self):
        if not self.recalculate_service_maps:
            return
        servicemapmanager.calculate_scheduled_service_maps_for_system(
            self.feed_update.feed.system
        )


class DirectionRuleSyncer(syncer(models.DirectionRule)):
    def sync(self, parsed_direction_rules):
        stop_id_to_pk = stopdam.get_id_to_pk_map_in_system(
            self.feed_update.feed.system.pk
        )
        entities_to_merge = []
        for parsed_direction_rule in parsed_direction_rules:
            direction_rule = models.DirectionRule.from_parsed_direction_rule(
                parsed_direction_rule
            )
            stop_pk = stop_id_to_pk.get(parsed_direction_rule.stop_id)
            if stop_pk is None:
                continue
            direction_rule.stop_pk = stop_pk
            entities_to_merge.append(direction_rule)

        __, num_added, num_updated = self._merge_entities(entities_to_merge)
        return num_added, num_updated

    def _get_id_to_pk_map(self):
        return genericqueries.get_id_to_pk_map_by_feed_pk(
            self.__db_entity__, self.feed_update.feed.pk
        )


@dataclasses.dataclass
class _TripStopTime(parse.TripStopTime):
    pk: int = None
    stop_pk: int = None
    trip_pk: int = None
    is_from_parsing: bool = True

    def is_new(self):
        return self.pk is None

    def to_db_mapping(self):
        result = dataclasses.asdict(self)
        del result["stop_id"]
        del result["is_from_parsing"]
        if not self.is_from_parsing:
            null_keys = {key for key, value in result.items() if value is None}
            for null_key in null_keys:
                del result[null_key]
        if self.is_new():
            del result["pk"]
        return result


@dataclasses.dataclass
class _Trip(parse.Trip):
    pk: int = None
    route_pk: int = None
    source_pk: int = None
    stop_times: typing.List[_TripStopTime] = dataclasses.field(default_factory=list)

    @classmethod
    def from_parsed_trip(cls, parsed_trip: parse.Trip) -> "_Trip":
        parsed_trip_dict = dataclasses.asdict(parsed_trip)
        parsed_stop_times_dict = parsed_trip_dict.pop("stop_times")
        return cls(
            **parsed_trip_dict,
            stop_times=[
                _TripStopTime(**parsed_stop_time_dict, is_from_parsing=True)
                for parsed_stop_time_dict in parsed_stop_times_dict
            ]
        )

    def is_new(self):
        return self.pk is None

    def to_db_mapping(self):
        result = {
            "pk": self.pk,
            "id": self.id,
            "route_pk": self.route_pk,
            "source_pk": self.source_pk,
            "direction_id": self.direction_id,
            "start_time": self.start_time,
            "last_update_time": self.updated_at,
            "vehicle_id": self.train_id,
            "current_status": self.current_status,
            "current_stop_sequence": self.current_stop_sequence,
        }
        if self.is_new():
            del result["pk"]
        return result


class TripSyncer(syncer(models.Trip)):

    route_pk_to_previous_service_map_hash = {}
    route_pk_to_new_service_map_hash = {}
    stop_time_pks_to_delete = set()

    def sync(self, parsed_trips):
        trips = map(_Trip.from_parsed_trip, parsed_trips)
        for data_adder in (
            self._add_source,
            self._add_schedule_data,  # Must come before route data
            self._add_route_data,
            self._add_stop_data,  # Must come before existing trip data
            self._add_existing_trip_data,
        ):
            trips = data_adder(trips)
        self._calculate_route_pk_to_new_service_map_hash(trips)
        return self._fast_merge(trips)

    def _add_source(self, trips: Iterable[_Trip]) -> Iterable[_Trip]:
        for trip in trips:
            trip.source_pk = self.feed_update.pk
            yield trip

    def _add_schedule_data(self, trips: Iterable[_Trip]) -> Iterable[_Trip]:
        """
        Add data to the trip, such as the route, from the schedule.
        """
        trips = list(trips)
        trip_ids_needing_schedule = set()
        for trip in trips:
            route_set = trip.route_id is not None
            direction_id_set = trip.direction_id is not None
            if route_set and direction_id_set:
                continue
            trip_ids_needing_schedule.add(trip.id)
        if len(trip_ids_needing_schedule) == 0:
            return trips
        trip_id_to_scheduled_trip = {
            trip.id: trip
            for trip in scheduledam.list_trips_by_system_pk_and_trip_ids(
                self.feed_update.feed.system.pk, trip_ids_needing_schedule
            )
        }
        for trip in trips:
            scheduled_trip = trip_id_to_scheduled_trip.get(trip.id)
            if scheduled_trip is None:
                continue
            trip.route_pk = scheduled_trip.route_pk
            if trip.direction_id is None:
                trip.direction_id = scheduled_trip.direction_id
        return trips

    def _add_route_data(self, trips: Iterable[_Trip]) -> Iterable[_Trip]:
        """
        Convert route_ids on the trip into route_pks. Trips that are have invalid
        route IDs and are missing route PKs are filtered out.
        """
        trips = list(trips)
        route_id_to_pk = routedam.get_id_to_pk_map_in_system(
            self.feed_update.feed.system.pk,
            [trip.route_id for trip in trips if trip.route_id is not None],
        )
        for trip in trips:
            if trip.route_pk is None:
                trip.route_pk = route_id_to_pk.get(trip.route_id)
                if trip.route_pk is None:
                    continue
            yield trip

    def _add_stop_data(self, trips: Iterable[_Trip]) -> Iterable[_Trip]:
        """
        Convert stop_ids on the trip stop times into stop_pks. Trip stop times that have
        invalid stop IDs and are missing stop PKs are filtered out.
        """
        stop_id_to_pk = stopdam.get_id_to_pk_map_in_system(
            self.feed_update.feed.system.pk
        )

        def process_stop_times(stop_times):
            for stop_time in stop_times:
                stop_pk = stop_id_to_pk.get(stop_time.stop_id, None)
                if stop_pk is None:
                    continue
                stop_time.stop_pk = stop_pk
                yield stop_time

        for trip in trips:
            trip.stop_times = list(process_stop_times(trip.stop_times))
            yield trip

    def _add_existing_trip_data(self, trips: Iterable[_Trip]) -> Iterable[_Trip]:
        """
        Add data to the feed trips from data already in the database; i.e., from
        previous feed updates.
        """
        trips = list(trips)
        trip_id_to_db_trip = self._build_trip_id_to_db_trip_map(
            trip.id for trip in trips
        )
        trip_pk_to_db_stop_time_data_list = tripdam.get_trip_pk_to_stop_time_data_list(
            self.feed_update.feed.pk
        )
        for trip in trips:
            db_trip = trip_id_to_db_trip.get(trip.id, None)
            if db_trip is None:
                continue
            trip.pk = db_trip.pk
            db_stop_time_data = trip_pk_to_db_stop_time_data_list.get(db_trip.pk, [])
            past_stop_times_iter = self._build_past_stop_times(
                trip, db_trip, db_stop_time_data
            )
            self._add_future_stop_time_data_to_trip(trip, db_stop_time_data)
            trip.stop_times = list(past_stop_times_iter) + trip.stop_times

        self._calculate_stop_time_pks_to_delete(
            trips, trip_pk_to_db_stop_time_data_list
        )
        self._calculate_route_pk_to_previous_service_map_hash(
            trip_id_to_db_trip.values(), trip_pk_to_db_stop_time_data_list
        )
        return trips

    def _build_trip_id_to_db_trip_map(self, feed_trip_ids):
        trip_id_to_db_trip = {
            trip.id: trip
            for trip in tripdam.list_all_from_feed(self.feed_update.feed.pk)
        }
        missing_trip_ids = set(
            trip_id for trip_id in feed_trip_ids if trip_id not in trip_id_to_db_trip
        )
        trip_id_to_db_trip.update(
            {
                trip.id: trip
                for trip in tripdam.list_by_system_and_trip_ids(
                    self.feed_update.feed.system.pk, missing_trip_ids
                )
            }
        )
        return trip_id_to_db_trip

    @staticmethod
    def _build_past_stop_times(
        trip, db_trip, db_stop_time_data_list: List[tripdam.StopTimeData]
    ) -> Iterable[_TripStopTime]:
        """
        Build stop times that have already passed and are not in the feed.
        """
        if db_trip is None:
            return
        if len(trip.stop_times) == 0:
            return
        first_future_stop_sequence = trip.stop_times[0].stop_sequence
        # Used to ensure that duplicate stops are not put into the DB. This is a
        # safety measure until #36 is resolved.
        future_stop_pks = set(stop_time.stop_pk for stop_time in trip.stop_times)
        for stop_time_data in db_stop_time_data_list:
            if stop_time_data.stop_sequence >= first_future_stop_sequence:
                return
            if stop_time_data.stop_pk in future_stop_pks:
                return
            past_stop_time = _TripStopTime(
                pk=stop_time_data.pk,
                stop_pk=stop_time_data.stop_pk,
                stop_sequence=stop_time_data.stop_sequence,
                future=False,
                stop_id="",
                is_from_parsing=False,
            )
            yield past_stop_time

    @staticmethod
    def _add_future_stop_time_data_to_trip(trip, db_stop_time_data):
        """
        Add stop time data from the database trip.
        """
        stop_sequence_to_pk_and_stop_pk = {
            stop_sequence: (stop_time_pk, stop_pk)
            for stop_time_pk, stop_sequence, stop_pk in db_stop_time_data
        }
        for feed_stop_time in trip.stop_times:
            stop_time_pk, __ = stop_sequence_to_pk_and_stop_pk.get(
                feed_stop_time.stop_sequence, (None, None)
            )
            if stop_time_pk is not None:
                feed_stop_time.pk = stop_time_pk

    def _calculate_stop_time_pks_to_delete(self, trips, trip_pk_to_db_stop_time_data):
        """
        Calculate the stop time pks to delete and store them in the object variable.
        """
        updated_stop_time_pks = set(
            stop_time.pk
            for trip in trips
            for stop_time in trip.stop_times
            if stop_time.pk is not None
        )
        self.stop_time_pks_to_delete = set(
            stop_time_data.pk
            for db_stop_time_data_list in trip_pk_to_db_stop_time_data.values()
            for stop_time_data in db_stop_time_data_list
            if stop_time_data.pk not in updated_stop_time_pks
        )

    def _calculate_route_pk_to_previous_service_map_hash(
        self, db_trips, trip_pk_to_db_stop_time_data_list
    ):
        """
        Calculate the previous service map information and store it in the object
        variable.
        """
        route_pk_to_trip_paths = collections.defaultdict(set)
        for db_trip in db_trips:
            db_stop_time_data_list = trip_pk_to_db_stop_time_data_list.get(
                db_trip.pk, []
            )
            if len(db_stop_time_data_list) == 0:
                continue
            if not db_trip.direction_id:
                db_stop_time_data_list = reversed(db_stop_time_data_list)
            route_pk_to_trip_paths[db_trip.route_pk].add(
                tuple(
                    stop_time_data.stop_pk for stop_time_data in db_stop_time_data_list
                )
            )
        self.route_pk_to_previous_service_map_hash = {
            route_pk: servicemapmanager.calculate_paths_hash(paths)
            for route_pk, paths in route_pk_to_trip_paths.items()
        }

    def _calculate_route_pk_to_new_service_map_hash(self, trips):
        """
        Calculate the new service map information and store it in the object
        variable.
        """
        route_pk_to_trip_paths = collections.defaultdict(set)
        for trip in trips:
            if len(trip.stop_times) == 0:
                continue
            if trip.direction_id:
                stop_times = trip.stop_times
            else:
                stop_times = reversed(trip.stop_times)
            route_pk_to_trip_paths[trip.route_pk].add(
                tuple(stop_time.stop_pk for stop_time in stop_times)
            )
        self.route_pk_to_new_service_map_hash = {
            route_pk: servicemapmanager.calculate_paths_hash(paths)
            for route_pk, paths in route_pk_to_trip_paths.items()
        }

    def _fast_merge(self, trips):
        num_added, num_updated = self._fast_mappings_merge(models.Trip, trips)
        trip_id_to_db_trip_pk = genericqueries.get_id_to_pk_map_by_feed_pk(
            models.Trip, self.feed_update.feed.pk
        )
        dbconnection.get_session().query(models.TripStopTime).filter(
            models.TripStopTime.pk.in_(self.stop_time_pks_to_delete)
        ).delete(synchronize_session=False)
        for trip in trips:
            trip_pk = trip_id_to_db_trip_pk[trip.id]
            for stop_time in trip.stop_times:
                stop_time.trip_pk = trip_pk
        self._fast_mappings_merge(
            models.TripStopTime,
            (stop_time for trip in trips for stop_time in trip.stop_times),
        )
        return num_added, num_updated

    @staticmethod
    def _fast_mappings_merge(db_model, entities):
        new_mappings = []
        updated_mappings = []
        for entity in entities:
            if entity.is_new():
                new_mappings.append(entity.to_db_mapping())
            else:
                updated_mappings.append(entity.to_db_mapping())
        session = dbconnection.get_session()
        session.bulk_insert_mappings(db_model, new_mappings)
        session.bulk_update_mappings(db_model, updated_mappings)
        session.flush()
        return len(new_mappings), len(updated_mappings)

    def post_sync(self):
        changed_route_pks = servicemapmanager.calculate_changed_route_pks_from_hashes(
            self.route_pk_to_previous_service_map_hash,
            self.route_pk_to_new_service_map_hash,
        )
        if len(changed_route_pks) == 0:
            return
        route_pk_to_route = {
            route.pk: route
            for route in routedam.list_all_in_system(self.feed_update.feed.system.id)
        }
        for route_pk in changed_route_pks:
            servicemapmanager.calculate_realtime_service_map_for_route(
                route_pk_to_route[route_pk]
            )


class AlertSyncer(syncer(models.Alert)):
    def sync(self, parsed_alerts):
        # NOTE: when working on this function as part of the full alerts support,
        # we should make it less efficient by not relying on ORM methods like
        # models.System.routes.
        persisted_alerts, num_added, num_updated = self._merge_entities(
            list(map(models.Alert.from_parsed_alert, parsed_alerts))
        )
        route_id_to_route = {
            route.id: route for route in self.feed_update.feed.system.routes
        }
        alert_id_to_route_ids = {alert.id: alert.route_ids for alert in parsed_alerts}
        alert_id_to_agency_ids = {alert.id: alert.agency_ids for alert in parsed_alerts}
        for alert in persisted_alerts:
            alert.routes = [
                route_id_to_route[route_id]
                for route_id in alert_id_to_route_ids.get(alert.id, [])
                if route_id in route_id_to_route
            ]
            # NOTE: this is a temporary thing pending the creation of models.Agency
            if len(alert_id_to_agency_ids.get(alert.id, [])) > 0:
                alert.system_pk = self.feed_update.feed.system.pk
        return num_added, num_updated



"""
Module containing all Exceptions that can be raised in Transiter.
"""
import traceback

import inflection


# NOTE: All exceptions in this module have two requirements:
#
# (1) They must be subclasses of _TransiterException.
#
# (2) They must have an associated HTTP TripStatus in the HTTP Manager.
#
# If an exception is added breaking one of these, one of the HTTP Manager
# unit tests will very intentionally fail.
#
# NOTE: (2) may potentially be revisited if there are exceptions that are
# never expected to make it to the HTTP layer. Another idea is to permit any
# ancestor of the exception to have a HTTP status.


class TransiterException(Exception):
    """
    Base exception for all Transiter exceptions.
    """

    code = None
    message = None
    additional_info = {}

    def response(self):
        """
        Return structured information about this exception instance.
        """
        if len(self.additional_info) > 0:
            additional_info = {"parameters": self.additional_info}
        else:
            additional_info = {}
        return {
            "type": inflection.underscore(type(self).__name__).upper(),
            "code": self.code,
            "message": self.message,
            **additional_info,
        }

    def __init__(self, message=None):
        if message is not None:
            self.message = message


class InstallError(TransiterException):
    """
    Exception that is thrown when there's a problem during install.
    """

    code = "T010"
    message = "There was an error installing the transit system."


class UnexpectedError(TransiterException):
    """
    Exception that is thrown when there's a problem during install.
    """

    code = "T011"
    message = (
        "There was an unexpected error. This generally indicates a bug in Transiter"
    )

    def __init__(self, exception=None):
        if exception is not None:
            self.additional_info = {
                "type": type(exception).__name__,
                "message": str(exception),
                # The following formatting makes the stack readable in a browser
                "stack_trace": traceback.format_exc().strip().split("\n"),
            }


class InvalidInput(TransiterException):
    """
    Exception that is thrown when the input to a service layer function is
    invalid. This is usually the result of an invalid user HTTP request.
    """

    code = "T020"
    message = "The request contained invalid input."


class InvalidSystemConfigFile(TransiterException):
    code = "T029"
    message = "The system config file is invalid"


class IdNotFoundError(TransiterException):
    """
    Exception that is raised when a specific DB entity could not be found.
    """

    code = "T050"
    message = "One of the requested entities could not be found."

    def __init__(self, entity_type=None, **kwargs):
        if entity_type is None:
            return
        self.message = f"The {entity_type.__name__.lower()} does not exist."
        self.additional_info = kwargs
        self.additional_info["entity_type"] = entity_type.__name__.lower()


class PageNotFound(TransiterException):
    """
    Exception that is raised when a page cannot be found.
    """

    code = "T051"

    def __init__(self, request_path):
        self.message = "No page found at path '{}'.".format(request_path)
        self.additional_info = {"request_path": request_path}


class MethodNotAllowed(TransiterException):
    """
    Exception that is raised when unsupported method is used.
    """

    code = "T052"
    message = "The requested method is not allowed for this endpoint"

    def __init__(self, request_method, request_path, allowed_methods):
        self.message = (
            "The method {} is not supported for path '{}'. "
            "The allowed methods for this path are: {}."
        ).format(request_method, request_path, ", ".join(allowed_methods))
        self.additional_info = {
            "request_method": request_method,
            "request_path": request_path,
            "allowed_methods": allowed_methods,
        }


class InvalidPermissionsLevelInRequest(TransiterException):
    """
    Raised when the HTTP requests contains an unknown permissions level.
    """

    code = "T060"

    def __init__(self, permissions_level_str=None, valid_levels=[]):
        self.message = "Unknown permissions level in the HTTP request."
        self.additional_info = {
            "request_permissions_level": permissions_level_str,
            "valid_permissions_levels": valid_levels,
        }


class AccessDenied(TransiterException):
    """
    Raised when the HTTP request has insufficient permissions.
    """

    code = "T061"

    def __init__(self, provided=None, required=None):
        self.message = "Insufficient permission to access this HTTP endpoint."
        self.additional_info = {
            "required_permissions_level": required,
            "request_permissions_level": provided,
        }


class InternalDocumentationMisconfigured(TransiterException):
    code = "T070"
    message = "The documentation is currently unavailable"



"""
Add system auto update enabled field.

Revision ID: 65146f234144
Revises: b521e2a9055a
Create Date: 2020-03-23 21:57:11.720810

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "65146f234144"
down_revision = "b521e2a9055a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "system",
        sa.Column(
            "auto_update_enabled", sa.Boolean(), nullable=False, server_default="True"
        ),
    )

    # For consistency, change the column name on the feed table
    op.alter_column(
        "feed", "auto_update_on", new_column_name="auto_update_enabled", nullable=False
    )


def downgrade():
    pass



"""
The system service is used to install, delete and retrieve information about
(transit) systems.
"""
import datetime
import json
import logging
import time
import typing
import uuid
from dataclasses import dataclass
from typing import Iterator

from transiter import models, exceptions, __metadata__
from transiter.data import dbconnection
from transiter.data.dams import feeddam, systemdam, genericqueries
from transiter.executor import celeryapp
from transiter.scheduler import client
from transiter.services import systemconfigreader, views
from transiter.services.update import updatemanager

logger = logging.getLogger(__name__)


@dbconnection.unit_of_work
def list_all() -> typing.List[views.System]:
    """
    List all installed systems.
    """
    return list(map(views.System.from_model, systemdam.list_all()))


@dbconnection.unit_of_work
def get_by_id(system_id) -> views.SystemLarge:
    """
    Get information on a specific system.
    """
    system = systemdam.get_by_id(system_id)
    if system is None:
        raise exceptions.IdNotFoundError(models.System, system_id=system_id)
    response = views.SystemLarge.from_model(system)
    if system.status != system.SystemStatus.ACTIVE:
        return response
    response.stops = views.StopsInSystem.from_model(
        system, systemdam.count_stops_in_system(system_id)
    )
    response.routes = views.RoutesInSystem.from_model(
        system, systemdam.count_routes_in_system(system_id)
    )
    response.feeds = views.FeedsInSystem.from_model(
        system, systemdam.count_feeds_in_system(system_id)
    )
    response.agency_alerts = list(
        map(
            views.AlertLarge.from_model,
            systemdam.list_all_alerts_associated_to_system(system.pk),
        )
    )
    return response


def set_auto_update_enabled(system_id, auto_update):
    with dbconnection.inline_unit_of_work():
        response = systemdam.set_auto_update_enabled(system_id, auto_update)
    client.refresh_tasks()
    return response


def install(system_id, config_str, extra_settings, config_source_url, sync=True):
    """
    Install a transit system synchronously; i.e., in the current thread.

    This is a no-op if the system is already installed.

    This method is designed to be called both in a sync HTTP install request and as
    the main install method called by the executor of a async HTTP install request.

    If the transit system does not have a record in the DB when this method is called
    (as in the sync case), the install will take place in a single unit of work. One
    consequence of this is that if the install failed there will be no record of it
    in the DB.

    Otherwise a number of unit of works will be used to update the systems's status as
    it progresses through the install
    """
    logger.info("Received system install request for id={}".format(system_id))
    system_update_pk = _create_system_update(
        system_id, config_str, extra_settings, config_source_url
    )
    sync_to_function = {
        True: _execute_system_update,
        False: _execute_system_update_async.delay,
    }
    sync_to_function[sync](system_update_pk)
    return system_update_pk


@dbconnection.unit_of_work
def get_update_by_id(system_update_id) -> views.SystemUpdate:
    system_update = systemdam.get_update_by_pk(system_update_id)
    if system_update is None:
        raise exceptions.IdNotFoundError(models.SystemUpdate, id=str(system_update_id))
    result = views.SystemUpdate.from_model(system_update)
    result.system = views.System.from_model(system_update.system)
    return result


@dbconnection.unit_of_work
def _create_system_update(system_id, config_str, extra_settings, config_source_url):
    system = systemdam.get_by_id(system_id)
    if system is not None:
        invalid_statuses_for_update = {
            models.System.SystemStatus.SCHEDULED,
            models.System.SystemStatus.INSTALLING,
            models.System.SystemStatus.DELETING,
        }
        if system.status in invalid_statuses_for_update:
            raise exceptions.InstallError(
                "Cannot install or update system with status '{}'".format(
                    system.status.name
                )
            )
        elif system.status == models.System.SystemStatus.INSTALL_FAILED:
            system.status = models.System.SystemStatus.SCHEDULED
    else:
        system = models.System(
            id=system_id, status=models.System.SystemStatus.SCHEDULED,
        )
        systemdam.create(system)

    update = models.SystemUpdate(
        system=system,
        status=models.SystemUpdate.Status.SCHEDULED,
        config_template=config_str,
        config_parameters=json.dumps(extra_settings, indent=2),
        config_source_url=config_source_url,
        transiter_version=__metadata__.__version__,
    )
    dbconnection.get_session().flush()
    return update.pk


@celeryapp.app.task
def _execute_system_update_async(system_update_pk):
    return _execute_system_update(system_update_pk)


def _execute_system_update(system_update_pk):
    context = _mark_update_started(system_update_pk)

    # Pause all other activity on this system to avoid database concurrency problems
    # during the large upcoming unit of works.
    client.refresh_tasks()

    try:
        feed_ids_to_update, feed_ids_to_delete = _install_system_configuration(
            system_update_pk
        )
        for feed_id in feed_ids_to_update:
            update_status, __ = updatemanager.execute_feed_update(
                feed_update_pk=updatemanager.create_feed_update(
                    context.system_id, feed_id
                )
            )
            if update_status != models.FeedUpdate.Status.SUCCESS:
                raise exceptions.InstallError(
                    message="Failed to update feed with id={}; reason: {}".format(
                        feed_id, update_status
                    )
                )
        for feed_id in feed_ids_to_delete:
            _delete_feed(context.system_id, feed_id)
        _mark_update_completed(
            context, models.SystemUpdate.Status.SUCCESS,
        )
    except exceptions.TransiterException as e:
        _mark_update_completed(
            context, models.SystemUpdate.Status.FAILED, json.dumps(e.response()),
        )
        logger.debug("Install or update of system failed", exc_info=True)
    except Exception as e:
        _mark_update_completed(
            context,
            models.SystemUpdate.Status.FAILED,
            json.dumps(exceptions.UnexpectedError(str(e)).response()),
        )
        logger.debug("Install or update of system failed", exc_info=True)

    client.refresh_tasks()


@dataclass
class _SystemUpdateContext:
    update_pk: int
    system_id: str
    prior_auto_update_setting: bool
    start_time: float


@dbconnection.unit_of_work
def _mark_update_started(system_update_pk) -> _SystemUpdateContext:
    system_update = systemdam.get_update_by_pk(system_update_pk)
    system_update.status = models.SystemUpdate.Status.IN_PROGRESS
    system = system_update.system
    prior_auto_update_setting = system.auto_update_enabled
    system.auto_update_enabled = False
    if system.status == models.System.SystemStatus.SCHEDULED:
        system.status = models.System.SystemStatus.INSTALLING
    return _SystemUpdateContext(
        update_pk=system_update_pk,
        system_id=system.id,
        prior_auto_update_setting=prior_auto_update_setting,
        start_time=time.time(),
    )


@dbconnection.unit_of_work
def _mark_update_completed(
    context: _SystemUpdateContext, final_status, final_status_message=None,
):
    """
    Set the status of a transit system, if that system exists in the DB.
    """
    system_update = systemdam.get_update_by_pk(context.update_pk)
    system_update.status = final_status
    system_update.status_message = final_status_message
    system_update.total_duration = time.time() - context.start_time
    system_update.completed_at = datetime.datetime.utcnow()
    system_update.system.auto_update_enabled = context.prior_auto_update_setting
    if system_update.system.status == models.System.SystemStatus.INSTALLING:
        if final_status == models.SystemUpdate.Status.SUCCESS:
            system_update.system.status = models.System.SystemStatus.ACTIVE
        else:
            system_update.system.status = models.System.SystemStatus.INSTALL_FAILED


@dbconnection.unit_of_work
def _install_system_configuration(system_update_pk):
    system_update = systemdam.get_update_by_pk(system_update_pk)
    extra_settings = json.loads(system_update.config_parameters)
    system_update.config = systemconfigreader.render_template(
        system_update.config_template, extra_settings
    )
    system_config = systemconfigreader.read(system_update.config, extra_settings)
    system = system_update.system

    # Service maps must come first in case calculations are triggered
    # by install_required feed updates
    _save_service_map_configuration(
        system, system_config[systemconfigreader.SERVICE_MAPS]
    )
    return _save_feed_configuration(system, system_config[systemconfigreader.FEEDS])


def delete_by_id(system_id, error_if_not_exists=True, sync=True):
    """
    Delete a transit system
    """
    with dbconnection.inline_unit_of_work():
        system = systemdam.get_by_id(system_id)
        if system is not None:
            system.status = models.System.SystemStatus.DELETING
            if not sync:
                system.id = system.id + "_deleting_" + str(uuid.uuid4())
                system_id = system.id
        elif error_if_not_exists:
            raise exceptions.IdNotFoundError
        else:
            return
    client.refresh_tasks()

    sync_to_function = {
        True: _complete_delete_operation,
        False: _complete_delete_operation_async.delay,
    }
    sync_to_function[sync](system_id)


@celeryapp.app.task
def _complete_delete_operation_async(system_id):
    return _complete_delete_operation(system_id)


def _complete_delete_operation(system_id):
    feed_ids = set()
    with dbconnection.inline_unit_of_work():
        system = systemdam.get_by_id(system_id)
        for feed in system.feeds:
            feed_ids.add(feed.id)

    for feed_id in feed_ids:
        _delete_feed(system_id, feed_id)

    with dbconnection.inline_unit_of_work():
        systemdam.delete_by_id(system_id)


def _delete_feed(system_id, feed_id):
    logger.info("Deleting feed {}/{}".format(system_id, feed_id))
    updatemanager.execute_feed_update(
        updatemanager.create_feed_flush(system_id, feed_id)
    )
    with dbconnection.inline_unit_of_work():
        feeddam.delete_in_system_by_id(system_id, feed_id)


def _save_feed_configuration(system, feeds_config):
    """
    Save feeds in a system.

    Stale feed maps -- those that are currently attached to the system but that
    do not correspond to entries in the config -- are *not* deleted by this method.

    :param system: the system to save the service maps in
    :param feeds_config: the feeds config JSON blob
    :return: a two-tuple of feed ID collections; the first collection contains IDs for
     feeds that need to be updated and the second for feeds that need to be deleted
    """
    feed_id_to_pk = genericqueries.get_id_to_pk_map(models.Feed, system.pk)
    session = dbconnection.get_session()
    feed_ids_to_update = list()
    for feed in _build_feeds_from_config(feeds_config):
        feed.system_pk = system.pk
        if feed.id in feed_id_to_pk:
            feed.pk = feed_id_to_pk[feed.id]
            del feed_id_to_pk[feed.id]
            logger.info("Updating feed {}/{}".format(system.id, feed.id))
        else:
            logger.info("Creating feed {}/{}".format(system.id, feed.id))
        session.merge(feed)
        if not feed.required_for_install:
            continue
        feed_ids_to_update.append(feed.id)

    feed_ids_to_delete = list(feed_id_to_pk.keys())
    return feed_ids_to_update, feed_ids_to_delete


def _build_feeds_from_config(feeds_config) -> Iterator[models.Feed]:
    for id_, config in feeds_config.items():
        feed = models.Feed()
        feed.id = id_
        feed.built_in_parser = config[systemconfigreader.PARSER].get(
            systemconfigreader.BUILT_IN
        )
        feed.custom_parser = config[systemconfigreader.PARSER].get(
            systemconfigreader.CUSTOM
        )
        feed.url = config[systemconfigreader.HTTP][systemconfigreader.URL]
        feed.headers = json.dumps(
            dict(config[systemconfigreader.HTTP][systemconfigreader.HEADERS])
        )
        feed.auto_update_enabled = config[systemconfigreader.AUTO_UPDATE][
            systemconfigreader.ENABLED
        ]
        feed.auto_update_period = config[systemconfigreader.AUTO_UPDATE][
            systemconfigreader.PERIOD
        ]
        feed.required_for_install = config[systemconfigreader.REQUIRED_FOR_INSTALL]
        yield feed


def _save_service_map_configuration(system, service_maps_config):
    """
    Save service maps in a system.

    Stale service maps -- those that are currently attached to the system but that
    do not correspond to entries in the config -- are deleted.

    :param system: the system to save the service maps in
    :param service_maps_config: the service maps config JSON blob
    """
    group_id_to_existing_service_map_group = {
        service_map_group.id: service_map_group
        for service_map_group in system.service_map_groups
    }
    session = dbconnection.get_session()
    for service_map_group in _build_service_maps_from_config(service_maps_config):
        service_map_group.system_pk = system.pk
        existing_service_map_group = group_id_to_existing_service_map_group.get(
            service_map_group.id
        )
        if existing_service_map_group is not None:
            service_map_group.pk = existing_service_map_group.pk
            del group_id_to_existing_service_map_group[service_map_group.id]
            logger.info(
                "Updating service map {}/{}".format(system.id, service_map_group.id)
            )
        else:
            logger.info(
                "Creating service map {}/{}".format(system.id, service_map_group.id)
            )
        session.merge(service_map_group)

    for existing_service_map_group in group_id_to_existing_service_map_group.values():
        logger.info(
            "Deleting service map {}/{}".format(
                system.id, existing_service_map_group.id
            )
        )
        session.delete(existing_service_map_group)


def _build_service_maps_from_config(
    service_maps_config,
) -> Iterator[models.ServiceMapGroup]:
    for id_, config in service_maps_config.items():
        group = models.ServiceMapGroup()
        group.id = id_
        group.source = config[systemconfigreader.SOURCE]
        json_conditions = config.get(systemconfigreader.CONDITIONS)
        if json_conditions is not None:
            group.conditions = json.dumps(json_conditions, indent=2)
        else:
            group.conditions = None
        group.threshold = config[systemconfigreader.THRESHOLD]
        group.use_for_routes_at_stop = config[systemconfigreader.USE_FOR_ROUTES_AT_STOP]
        group.use_for_stops_in_route = config[systemconfigreader.USE_FOR_STOPS_IN_ROUTE]
        yield group



from .adminendpoints import admin_endpoints
from .docsendpoints import docs_endpoints
from .feedendpoints import feed_endpoints
from .routeendpoints import route_endpoints
from .stopendpoints import stop_endpoints
from .systemendpoints import system_endpoints
from .tripendpoints import trip_endpoints



import flask

from transiter.http.httpmanager import http_endpoint, link_target
from transiter.services import routeservice, views

route_endpoints = flask.Blueprint(__name__, __name__)


@http_endpoint(route_endpoints, "")
@link_target(views.RoutesInSystem, ["_system_id"])
def list_all_in_system(system_id):
    """List all routes for a specific system."""
    return routeservice.list_all_in_system(system_id)


@http_endpoint(route_endpoints, "/<route_id>")
@link_target(views.Route, ["_system_id", "id"])
def get_in_system_by_id(system_id, route_id):
    """Retrieve a specific route in a specific system."""
    return routeservice.get_in_system_by_id(system_id, route_id)



from sqlalchemy import Column, Integer, String, ForeignKey, Enum, UniqueConstraint
from sqlalchemy.orm import relationship

from transiter import parse
from .base import Base
from .updatableentity import updatable_from


@updatable_from(parse.Route)
class Route(Base):
    __tablename__ = "route"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    system_pk = Column(Integer, ForeignKey("system.pk"), nullable=False)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    Type = parse.Route.Type

    color = Column(String)
    text_color = Column(String)
    short_name = Column(String)
    long_name = Column(String)
    description = Column(String)
    url = Column(String)
    type = Column(Enum(Type, name="route_type", native_enum=False))
    sort_order = Column(Integer)

    system = relationship("System", back_populates="routes")
    source = relationship("FeedUpdate", cascade="none")
    trips = relationship("Trip", back_populates="route", cascade="all, delete-orphan")
    scheduled_trips = relationship(
        "ScheduledTrip", back_populates="route", cascade="all, delete-orphan"
    )
    service_maps = relationship(
        "ServiceMap",
        back_populates="route",
        primaryjoin="ServiceMap.route_pk==Route.pk",
        cascade="all, delete-orphan",
    )
    route_statuses = relationship(
        "Alert", secondary="alert_route", back_populates="routes", cascade="all"
    )

    __table_args__ = (UniqueConstraint("system_pk", "id"),)

    @staticmethod
    def from_parsed_route(route: parse.Route) -> "Route":
        return Route(
            id=route.id,
            type=route.type,
            short_name=route.short_name,
            long_name=route.long_name,
            description=route.description,
            color=route.color,
            text_color=route.text_color,
            url=route.url,
            sort_order=route.sort_order,
        )



"""
This module contains the code for the transiterclt (Transiter command line
tools) command line program.
"""
import time

import click
from sqlalchemy import exc

from transiter import config
from transiter.data import dbconnection
from transiter.executor import celeryapp
from transiter.scheduler import server as scheduler, client


@click.group()
def transiter_clt():
    """
    Transiter Command Line Tools
    """
    pass


@transiter_clt.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force start by killing any process listening on the target port.",
)
@click.option(
    "-l",
    "--log-level",
    default="warning",
    type=click.Choice(["debug", "info", "warning", "error", "fatal"]),
    help="The log level.",
)
@click.argument("server", type=click.Choice(["webservice", "scheduler", "executor"]))
def launch(force, log_level, server):
    """
    Launch a Transiter service in debug mode.
    """
    if server == "webservice":
        # NOTE: the flask app is imported here because otherwise the task server will
        # use the app's logging configuration.
        from transiter.http import flaskapp

        flaskapp.launch(force)
    if server == "scheduler":
        app = scheduler.create_app()
        app.run(host="0.0.0.0", port=config.SCHEDULER_PORT, debug=False)
    if server == "executor":
        celeryapp.run(log_level)


@transiter_clt.group()
def db():
    """
    Perform Transiter database operation.
    """


@db.command()
def init():
    """
    Initialize the Transiter database.

    This is a no-op if the database is already initialized to any version of the
    database. It is designed to safely avoid accidental upgrades.
    """
    while True:
        try:
            if dbconnection.get_current_database_revision() is not None:
                print(
                    "The database has already been initialized and "
                    "no upgrade will be attempted."
                )
                print(
                    "In order to upgrade the database use the "
                    "`transiterclt db upgrade` command. "
                )
                return
            print("Initializing the database")
            dbconnection.upgrade_database()
            print("Database initialized!")
        except exc.SQLAlchemyError:
            print("Failed to connect to the database; Waiting 1 second")
            time.sleep(1)


@db.command()
def upgrade():
    """
    Upgrade the Transiter database to its latest version.
    """
    dbconnection.upgrade_database()


@db.command()
@click.confirmation_option(
    prompt="This will result in all data in the current database being lost.\n"
    "Are you sure?"
)
def reset():
    """
    Reset the Transiter database.

    This operation drops all of the Transiter tables in the database, if they
    exist, and then creates them. All existing data will be lost.
    """
    dbconnection.delete_all_tables()
    dbconnection.upgrade_database()
    client.refresh_tasks()



from alembic import context

from transiter import models
from transiter.data import dbconnection

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

target_metadata = models.Base.metadata


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = dbconnection.create_engine()

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


run_migrations_online()



from sqlalchemy import Column, UniqueConstraint, Index, Time, Integer, ForeignKey
from sqlalchemy.orm import relationship

from .base import Base


class ScheduledTripStopTime(Base):
    __tablename__ = "scheduled_trip_stop_time"

    pk = Column(Integer, primary_key=True)
    trip_pk = Column(Integer, ForeignKey("scheduled_trip.pk"), nullable=False)
    stop_pk = Column(Integer, ForeignKey("stop.pk"), nullable=False)

    arrival_time = Column(Time(timezone=False))
    departure_time = Column(Time(timezone=False))
    stop_sequence = Column(Integer, nullable=False)

    stop = relationship("Stop", back_populates="scheduled_trip_times", cascade="")
    trip = relationship("ScheduledTrip", back_populates="stop_times", cascade="")

    __table_args__ = (
        UniqueConstraint("trip_pk", "stop_sequence"),
        Index(
            "scheduled_trip_stop_time_trip_pk_departure_time_idx",
            "trip_pk",
            "departure_time",
        ),
    )



ALL_OF = "all_of"
DAYS = "days"
ENDS_EARLIER_THAN = "ends_earlier_than"
ENDS_LATER_THAN = "ends_later_than"
NONE_OF = "none_of"
ONE_OF = "one_of"
STARTS_EARLIER_THAN = "starts_earlier_than"
STARTS_LATER_THAN = "starts_later_than"
WEEKDAY = "weekday"
WEEKEND = "weekend"
